# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.authentication import (
    AWSEmbeddingHeadersProvider,
    EmbeddingAPIKeyHeaderProvider,
    coerce_embedding_headers_provider,
)
from astrapy.settings.defaults import (
    EMBEDDING_HEADER_API_KEY,
    EMBEDDING_HEADER_AWS_ACCESS_ID,
    EMBEDDING_HEADER_AWS_SECRET_ID,
)


class TestEmbeddingHeadersProvider:
    @pytest.mark.describe("test of headers from EmbeddingAPIKeyHeaderProvider")
    def test_embeddingheadersprovider_static(self) -> None:
        ehp = EmbeddingAPIKeyHeaderProvider("x")
        assert {k.lower(): v for k, v in ehp.get_headers().items()} == {
            EMBEDDING_HEADER_API_KEY.lower(): "x"
        }

    @pytest.mark.describe("test of headers from empty EmbeddingAPIKeyHeaderProvider")
    def test_embeddingheadersprovider_null(self) -> None:
        ehp = EmbeddingAPIKeyHeaderProvider(None)
        assert ehp.get_headers() == {}

    @pytest.mark.describe("test of headers from AWSEmbeddingHeadersProvider")
    def test_embeddingheadersprovider_aws(self) -> None:
        ehp = AWSEmbeddingHeadersProvider(
            embedding_access_id="x",
            embedding_secret_id="y",
        )
        gen_headers_lower = {k.lower(): v for k, v in ehp.get_headers().items()}
        exp_headers_lower = {
            EMBEDDING_HEADER_AWS_ACCESS_ID.lower(): "x",
            EMBEDDING_HEADER_AWS_SECRET_ID.lower(): "y",
        }
        assert gen_headers_lower == exp_headers_lower

    @pytest.mark.describe("test of embedding headers provider coercion")
    def test_embeddingheadersprovider_coercion(self) -> None:
        """This doubles as equality test."""
        ehp_s = EmbeddingAPIKeyHeaderProvider("x")
        ehp_n = EmbeddingAPIKeyHeaderProvider(None)
        ehp_a = AWSEmbeddingHeadersProvider(
            embedding_access_id="x",
            embedding_secret_id="y",
        )
        assert coerce_embedding_headers_provider(ehp_s) == ehp_s
        assert coerce_embedding_headers_provider(ehp_n) == ehp_n
        assert coerce_embedding_headers_provider(ehp_a) == ehp_a

        assert coerce_embedding_headers_provider("x") == ehp_s



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import time
from typing import TYPE_CHECKING

import pytest

from astrapy import Database
from astrapy.exceptions import DataAPIResponseException
from astrapy.ids import UUID

from ..conftest import CQL_AVAILABLE
from .table_cql_assets import (
    CREATE_TABLE_COUNTER,
    CREATE_TABLE_LOWSUPPORT,
    DROP_TABLE_COUNTER,
    DROP_TABLE_LOWSUPPORT,
    EXPECTED_ROW_COUNTER,
    EXPECTED_ROW_LOWSUPPORT,
    FILTER_COUNTER,
    FILTER_LOWSUPPORT,
    ILLEGAL_PROJECTIONS_LOWSUPPORT,
    INSERTS_TABLE_COUNTER,
    INSERTS_TABLE_LOWSUPPORT,
    LOWSUPPORT_TIMEUUID_DOC0,
    LOWSUPPORT_TIMEUUID_DOC1,
    LOWSUPPORT_TIMEUUID_PK0,
    LOWSUPPORT_TIMEUUID_PK1,
    PROJECTION_LOWSUPPORT,
    TABLE_NAME_COUNTER,
    TABLE_NAME_LOWSUPPORT,
)

if TYPE_CHECKING:
    from cassandra.cluster import Session

try:
    from cassandra.cluster import Session
except ImportError:
    pass


@pytest.mark.skipif(not CQL_AVAILABLE, reason="No CQL session available")
class TestTableCQLDrivenDMLSync:
    @pytest.mark.describe(
        "test of reading from a CQL-driven table with a Counter, sync"
    )
    def test_table_cqldriven_counter_sync(
        self,
        cql_session: Session,
        sync_database: Database,
    ) -> None:
        try:
            cql_session.execute(CREATE_TABLE_COUNTER)
            for insert_statement in INSERTS_TABLE_COUNTER:
                cql_session.execute(insert_statement)
            time.sleep(1.5)  # delay for schema propagation

            table = sync_database.get_table(TABLE_NAME_COUNTER)
            table.definition()
            row = table.find_one(filter=FILTER_COUNTER)
            assert row == EXPECTED_ROW_COUNTER
            table.delete_one(filter=FILTER_COUNTER)
            row = table.find_one(filter=FILTER_COUNTER)
            assert row is None
        finally:
            cql_session.execute(DROP_TABLE_COUNTER)

    @pytest.mark.describe(
        "test of reading from a CQL-driven table with limited-support columns, sync"
    )
    def test_table_cqldriven_lowsupport_sync(
        self,
        cql_session: Session,
        sync_database: Database,
    ) -> None:
        try:
            cql_session.execute(CREATE_TABLE_LOWSUPPORT)
            for insert_statement in INSERTS_TABLE_LOWSUPPORT:
                cql_session.execute(insert_statement)
            time.sleep(1.5)  # delay for schema propagation

            table = sync_database.get_table(TABLE_NAME_LOWSUPPORT)
            table.definition()
            for ill_proj in ILLEGAL_PROJECTIONS_LOWSUPPORT:
                with pytest.raises(DataAPIResponseException):
                    table.find_one(filter=FILTER_LOWSUPPORT)
            row = table.find_one(
                filter=FILTER_LOWSUPPORT, projection=PROJECTION_LOWSUPPORT
            )
            assert row == EXPECTED_ROW_LOWSUPPORT
            table.delete_one(filter=FILTER_LOWSUPPORT)
            row = table.find_one(
                filter=FILTER_LOWSUPPORT, projection=PROJECTION_LOWSUPPORT
            )
            assert row is None

            # writing timeuuid, passing strings and an UUID object to insert_one
            table.insert_one(LOWSUPPORT_TIMEUUID_DOC0)
            r_timeuuid_doc0 = table.find_one(
                filter=LOWSUPPORT_TIMEUUID_PK0,
                projection={col: True for col in LOWSUPPORT_TIMEUUID_DOC0},
            )
            assert r_timeuuid_doc0 == LOWSUPPORT_TIMEUUID_DOC0
            table.insert_one(LOWSUPPORT_TIMEUUID_DOC1)
            r_timeuuid_doc1 = table.find_one(
                filter=LOWSUPPORT_TIMEUUID_PK1,
                projection={col: True for col in LOWSUPPORT_TIMEUUID_DOC1},
            )
            # the UUID form is expected when reading:
            assert r_timeuuid_doc1 is not None
            assert r_timeuuid_doc1["col_timeuuid"] == UUID(
                LOWSUPPORT_TIMEUUID_DOC1["col_timeuuid"]
            )
        finally:
            cql_session.execute(DROP_TABLE_LOWSUPPORT)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.constants import DefaultRowType
from astrapy.data.utils.table_converters import _TableConverterAgent
from astrapy.data_types import DataAPIDate
from astrapy.utils.api_options import defaultSerdesOptions


class TestTableConverterAgent:
    @pytest.mark.describe("test of table converter agent, row converters")
    def test_tableconverteragent_row(self) -> None:
        agent: _TableConverterAgent[DefaultRowType] = _TableConverterAgent(
            options=defaultSerdesOptions,
        )
        assert len(agent.row_postprocessors) == 0

        # try row of type 1
        schema1 = {
            "p_text": {"type": "text"},
            "p_boolean": {"type": "boolean"},
            "p_float_pinf": {"type": "float"},
        }
        raw_row1 = {
            "p_text": "italy",
            "p_boolean": True,
            "p_float_pinf": "Infinity",
        }
        agent.postprocess_row(
            raw_row1, columns_dict=schema1, similarity_pseudocolumn=None
        )
        assert len(agent.row_postprocessors) == 1

        # try row of type 2
        schema2 = {
            "p_float_pinf": {"type": "float"},
        }
        raw_row2 = {
            "p_float_pinf": "Infinity",
        }
        agent.postprocess_row(
            raw_row2, columns_dict=schema2, similarity_pseudocolumn=None
        )
        assert len(agent.row_postprocessors) == 2

        # try row of type 1 again
        raw_row1b = {
            "p_text": "croatia",
            "p_boolean": False,
            "p_float_pinf": 54.321,
        }
        agent.postprocess_row(
            raw_row1b, columns_dict=schema1, similarity_pseudocolumn=None
        )
        assert len(agent.row_postprocessors) == 2

    @pytest.mark.describe("test of table converter agent, row with similarity")
    def test_tableconverteragent_row_similarity(self) -> None:
        agent: _TableConverterAgent[DefaultRowType] = _TableConverterAgent(
            options=defaultSerdesOptions,
        )
        schema1 = {
            "$similarity": {"type": "date"},
        }

        raw_row1 = {
            "$similarity": 0.123,
        }
        result1 = agent.postprocess_row(
            raw_row1, columns_dict=schema1, similarity_pseudocolumn="$similarity"
        )
        assert isinstance(result1["$similarity"], float)

        raw_row2 = {
            "$similarity": "2021-11-11",
        }
        result2 = agent.postprocess_row(
            raw_row2, columns_dict=schema1, similarity_pseudocolumn=None
        )
        assert isinstance(result2["$similarity"], DataAPIDate)

    @pytest.mark.describe("test of table converter agent, key converters")
    def test_tableconverteragent_key(self) -> None:
        agent: _TableConverterAgent[DefaultRowType] = _TableConverterAgent(
            options=defaultSerdesOptions,
        )
        assert len(agent.key_postprocessors) == 0

        # try row of type 1
        k_schema1 = {
            "p_text": {"type": "text"},
            "p_boolean": {"type": "boolean"},
            "p_float_pinf": {"type": "float"},
        }
        pk_list1 = ["italy", True, "Infinity"]
        agent.postprocess_key(pk_list1, primary_key_schema_dict=k_schema1)
        assert len(agent.key_postprocessors) == 1

        # try row of type 2
        k_schema2 = {
            "p_float_pinf": {"type": "float"},
        }
        pk_list2 = ["Infinity"]
        agent.postprocess_key(pk_list2, primary_key_schema_dict=k_schema2)
        assert len(agent.key_postprocessors) == 2

        # try row of type 1 again
        pk_list1b = ["croatia", False, 54.321]
        agent.postprocess_key(pk_list1b, primary_key_schema_dict=k_schema1)
        assert len(agent.key_postprocessors) == 2



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import time

import pytest

from astrapy.exceptions import (
    DataAPITimeoutException,
    DevOpsAPITimeoutException,
    MultiCallTimeoutManager,
)


class TestTimeouts:
    @pytest.mark.describe("test MultiCallTimeoutManager")
    def test_multicalltimeoutmanager(self) -> None:
        mgr_n = MultiCallTimeoutManager(overall_timeout_ms=None)
        assert mgr_n.remaining_timeout().request_ms is None
        time.sleep(0.5)
        assert mgr_n.remaining_timeout().request_ms is None

        mgr_1 = MultiCallTimeoutManager(overall_timeout_ms=1000)
        crt_1 = mgr_1.remaining_timeout().request_ms
        assert crt_1 is not None
        time.sleep(0.6)
        crt_2 = mgr_1.remaining_timeout().request_ms
        assert crt_2 is not None
        time.sleep(0.6)
        with pytest.raises(DataAPITimeoutException):
            mgr_1.remaining_timeout().request_ms

    @pytest.mark.describe("test MultiCallTimeoutManager DevOps")
    def test_multicalltimeoutmanager_devops(self) -> None:
        mgr_n = MultiCallTimeoutManager(overall_timeout_ms=None, dev_ops_api=True)
        assert mgr_n.remaining_timeout().request_ms is None
        time.sleep(0.5)
        assert mgr_n.remaining_timeout().request_ms is None

        mgr_1 = MultiCallTimeoutManager(overall_timeout_ms=1000, dev_ops_api=True)
        crt_1 = mgr_1.remaining_timeout().request_ms
        assert crt_1 is not None
        time.sleep(0.6)
        crt_2 = mgr_1.remaining_timeout().request_ms
        assert crt_2 is not None
        time.sleep(0.6)
        with pytest.raises(DevOpsAPITimeoutException):
            mgr_1.remaining_timeout().request_ms



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any

from astrapy.constants import SortMode
from astrapy.info import (
    ColumnType,
    CreateTableDefinition,
    CreateTypeDefinition,
    TableKeyValuedColumnTypeDescriptor,
    TablePrimaryKeyDescriptor,
    TableScalarColumnTypeDescriptor,
    TableValuedColumnTypeDescriptor,
    TableVectorColumnTypeDescriptor,
    TableVectorIndexOptions,
    VectorServiceOptions,
)

TEST_ALL_RETURNS_TABLE_NAME = "test_table_all_returns"
TEST_ALL_RETURNS_TABLE_DEFINITION = CreateTableDefinition(
    columns={
        "p_ascii": TableScalarColumnTypeDescriptor(column_type="ascii"),
        "p_bigint": TableScalarColumnTypeDescriptor(column_type="bigint"),
        "p_blob": TableScalarColumnTypeDescriptor(column_type="blob"),
        "p_boolean": TableScalarColumnTypeDescriptor(column_type="boolean"),
        "p_date": TableScalarColumnTypeDescriptor(column_type="date"),
        "p_decimal": TableScalarColumnTypeDescriptor(column_type="decimal"),
        "p_double": TableScalarColumnTypeDescriptor(column_type="double"),
        "p_duration": TableScalarColumnTypeDescriptor(column_type="duration"),
        "p_float": TableScalarColumnTypeDescriptor(column_type="float"),
        "p_inet": TableScalarColumnTypeDescriptor(column_type="inet"),
        "p_int": TableScalarColumnTypeDescriptor(column_type="int"),
        "p_smallint": TableScalarColumnTypeDescriptor(column_type="smallint"),
        "p_text": TableScalarColumnTypeDescriptor(column_type="text"),
        "p_text_nulled": TableScalarColumnTypeDescriptor(column_type="text"),
        "p_text_omitted": TableScalarColumnTypeDescriptor(column_type="text"),
        "p_time": TableScalarColumnTypeDescriptor(column_type="time"),
        "p_timestamp": TableScalarColumnTypeDescriptor(column_type="timestamp"),
        "p_tinyint": TableScalarColumnTypeDescriptor(column_type="tinyint"),
        "p_varint": TableScalarColumnTypeDescriptor(column_type="varint"),
        "p_uuid": TableScalarColumnTypeDescriptor(column_type="uuid"),
        "p_vector": TableVectorColumnTypeDescriptor(
            column_type="vector",
            dimension=3,
            service=None,
        ),
        "p_list_int": TableValuedColumnTypeDescriptor(
            column_type="list",
            value_type="int",
        ),
        "p_map_text_text": TableKeyValuedColumnTypeDescriptor(
            column_type="map", key_type="text", value_type="text"
        ),
        "p_set_int": TableValuedColumnTypeDescriptor(
            column_type="set",
            value_type="int",
        ),
        "p_double_minf": TableScalarColumnTypeDescriptor(column_type="double"),
        "p_double_pinf": TableScalarColumnTypeDescriptor(column_type="double"),
        "p_float_nan": TableScalarColumnTypeDescriptor(column_type="float"),
    },
    primary_key=TablePrimaryKeyDescriptor(
        partition_by=["p_ascii", "p_bigint"],
        partition_sort={
            "p_int": SortMode.ASCENDING,
            "p_boolean": SortMode.DESCENDING,
        },
    ),
)

TEST_SIMPLE_TABLE_NAME = "test_table_simple"
TEST_SIMPLE_TABLE_DEFINITION = CreateTableDefinition(
    columns={
        "p_text": TableScalarColumnTypeDescriptor(column_type="text"),
        "p_int": TableScalarColumnTypeDescriptor(column_type="int"),
        "p_vector": TableVectorColumnTypeDescriptor(
            column_type="vector",
            dimension=3,
            service=None,
        ),
    },
    primary_key=TablePrimaryKeyDescriptor(
        partition_by=["p_text"],
        partition_sort={},
    ),
)
TEST_SIMPLE_TABLE_VECTOR_INDEX_NAME = "test_table_simple_p_vector_idx"
TEST_SIMPLE_TABLE_VECTOR_INDEX_COLUMN = "p_vector"
TEST_SIMPLE_TABLE_VECTOR_INDEX_OPTIONS = TableVectorIndexOptions(
    metric="cosine",
)

TEST_COMPOSITE_TABLE_NAME = "test_table_composite"
TEST_COMPOSITE_TABLE_DEFINITION = CreateTableDefinition(
    columns={
        "p_text": TableScalarColumnTypeDescriptor(column_type="text"),
        "p_int": TableScalarColumnTypeDescriptor(column_type="int"),
        "p_boolean": TableScalarColumnTypeDescriptor(column_type="boolean"),
        "p_vector": TableVectorColumnTypeDescriptor(
            column_type="vector",
            dimension=3,
            service=None,
        ),
    },
    primary_key=TablePrimaryKeyDescriptor(
        partition_by=["p_text"],
        partition_sort={"p_int": SortMode.ASCENDING},
    ),
)
TEST_COMPOSITE_TABLE_VECTOR_INDEX_NAME = "test_table_composite_p_vector_idx"
TEST_COMPOSITE_TABLE_VECTOR_INDEX_COLUMN = TEST_SIMPLE_TABLE_VECTOR_INDEX_COLUMN
TEST_COMPOSITE_TABLE_VECTOR_INDEX_OPTIONS = TEST_SIMPLE_TABLE_VECTOR_INDEX_OPTIONS
TEST_COMPOSITE_TABLE_BOOLEAN_INDEX_NAME = "test_table_composite_p_boolean_idx"
TEST_COMPOSITE_TABLE_BOOLEAN_INDEX_COLUMN = "p_boolean"
TEST_COMPOSITE_TABLE_BOOLEAN_INDEX_OPTIONS = None

TEST_ALLMAPS_TABLE_NAME = "test_table_allmaps"
TEST_ALLMAPS_TABLE_DEFINITION = (
    CreateTableDefinition.builder()
    .add_column("id", ColumnType.TEXT)
    .add_map_column("ascii_map", ColumnType.ASCII, ColumnType.ASCII)
    .add_map_column("bigint_map", ColumnType.BIGINT, ColumnType.BIGINT)
    .add_map_column("blob_map", ColumnType.BLOB, ColumnType.BLOB)
    .add_map_column("boolean_map", ColumnType.BOOLEAN, ColumnType.BOOLEAN)
    .add_map_column("date_map", ColumnType.DATE, ColumnType.DATE)
    .add_map_column("decimal_map", ColumnType.DECIMAL, ColumnType.DECIMAL)
    .add_map_column("double_map", ColumnType.DOUBLE, ColumnType.DOUBLE)
    .add_map_column("duration_map", ColumnType.TEXT, ColumnType.DURATION)
    .add_map_column("float_map", ColumnType.FLOAT, ColumnType.FLOAT)
    .add_map_column("inet_map", ColumnType.INET, ColumnType.INET)
    .add_map_column("int_map", ColumnType.INT, ColumnType.INT)
    .add_map_column("smallint_map", ColumnType.SMALLINT, ColumnType.SMALLINT)
    .add_map_column("text_map", ColumnType.TEXT, ColumnType.TEXT)
    .add_map_column("time_map", ColumnType.TIME, ColumnType.TIME)
    .add_map_column("timestamp_map", ColumnType.TIMESTAMP, ColumnType.TIMESTAMP)
    .add_map_column("tinyint_map", ColumnType.TINYINT, ColumnType.TINYINT)
    .add_map_column("uuid_map", ColumnType.UUID, ColumnType.UUID)
    .add_map_column("varint_map", ColumnType.VARINT, ColumnType.VARINT)
    .add_partition_by("id")
    .build()
)

TEST_VECTORIZE_TABLE_NAME = "test_table_vectorize"
TEST_VECTORIZE_TABLE_DEFINITION = CreateTableDefinition(
    columns={
        "p_text": TableScalarColumnTypeDescriptor(column_type="text"),
        "p_vector": TableVectorColumnTypeDescriptor(
            column_type="vector",
            dimension=64,
            service=VectorServiceOptions(
                provider="openai",
                model_name="text-embedding-3-small",
            ),
        ),
    },
    primary_key=TablePrimaryKeyDescriptor(
        partition_by=["p_text"],
        partition_sort={},
    ),
)
TEST_VECTORIZE_TABLE_VECTOR_INDEX_NAME = "test_table_vectorize_p_vector_idx"
TEST_VECTORIZE_TABLE_VECTOR_INDEX_COLUMN = TEST_SIMPLE_TABLE_VECTOR_INDEX_COLUMN
TEST_VECTORIZE_TABLE_VECTOR_INDEX_OPTIONS = TEST_SIMPLE_TABLE_VECTOR_INDEX_OPTIONS

TEST_KMS_VECTORIZE_TABLE_NAME = "test_table_kms_vectorize"
TEST_KMS_VECTORIZE_TABLE_DEFINITION = CreateTableDefinition(
    columns={
        "p_text": TableScalarColumnTypeDescriptor(column_type="text"),
        "p_vector": TableVectorColumnTypeDescriptor(
            column_type="vector",
            dimension=64,
            service=VectorServiceOptions(
                provider="openai",
                model_name="text-embedding-3-small",
                authentication={
                    "providerKey": "SHARED_SECRET_EMBEDDING_API_KEY_OPENAI"
                },
            ),
        ),
    },
    primary_key=TablePrimaryKeyDescriptor(
        partition_by=["p_text"],
        partition_sort={},
    ),
)
TEST_KMS_VECTORIZE_TABLE_VECTOR_INDEX_NAME = "test_table_kms_vectorize_p_vector_idx"
TEST_KMS_VECTORIZE_TABLE_VECTOR_INDEX_COLUMN = TEST_SIMPLE_TABLE_VECTOR_INDEX_COLUMN
TEST_KMS_VECTORIZE_TABLE_VECTOR_INDEX_DEFINITION = (
    TEST_SIMPLE_TABLE_VECTOR_INDEX_OPTIONS
)

TEST_MULTIPLEVECTORIZE_TABLE_NAME = "test_table_header_multiplevectorize"
TEST_MULTIPLEVECTORIZE_TABLE_DEFINITION = CreateTableDefinition(
    columns={
        "id": TableScalarColumnTypeDescriptor(column_type="text"),
        "p_text": TableScalarColumnTypeDescriptor(column_type="text"),
        "p_vector_small": TableVectorColumnTypeDescriptor(
            column_type="vector",
            dimension=1024,
            service=VectorServiceOptions(
                provider="openai",
                model_name="text-embedding-3-small",
            ),
        ),
        "p_vector_large": TableVectorColumnTypeDescriptor(
            column_type="vector",
            dimension=1024,
            service=VectorServiceOptions(
                provider="openai",
                model_name="text-embedding-3-large",
            ),
        ),
    },
    primary_key=TablePrimaryKeyDescriptor(
        partition_by=["id"],
        partition_sort={},
    ),
)
TEST_MULTIPLEVECTORIZE_TABLE_INDEXES: list[tuple[str, str]] = [
    ("test_idx_p_vector_small", "p_vector_small"),
    ("test_idx_p_vector_large", "p_vector_large"),
]
TEST_MULTIPLEVECTORIZE_TABLE_INDEX_OPTIONS = TEST_SIMPLE_TABLE_VECTOR_INDEX_OPTIONS

TEST_SIMPLE_UDT_NAME = "test_simple_udt"
TEST_SIMPLE_UDT_DEFINITION = CreateTypeDefinition(
    fields={
        "udt_text": TableScalarColumnTypeDescriptor(column_type=ColumnType.TEXT),
        "udt_int": TableScalarColumnTypeDescriptor(column_type=ColumnType.INT),
    }
)

TEST_COLLINDEXED_TABLE_NAME = "test_table_collindexed"
TEST_COLLINDEXED_TABLE_DEFINITION = CreateTableDefinition(
    columns={
        "id": TableScalarColumnTypeDescriptor(column_type="text"),
        "set_int": TableValuedColumnTypeDescriptor(
            column_type="set",
            value_type="int",
        ),
        "list_int": TableValuedColumnTypeDescriptor(
            column_type="list",
            value_type="int",
        ),
        "map_text_int_e": TableKeyValuedColumnTypeDescriptor(
            key_type="text",
            value_type="int",
        ),
        "map_text_int_k": TableKeyValuedColumnTypeDescriptor(
            key_type="text",
            value_type="int",
        ),
        "map_text_int_v": TableKeyValuedColumnTypeDescriptor(
            key_type="text",
            value_type="int",
        ),
        "map_int_text": TableKeyValuedColumnTypeDescriptor(
            key_type="int",
            value_type="text",
        ),
    },
    primary_key=TablePrimaryKeyDescriptor(
        partition_by=["id"],
        partition_sort={},
    ),
)
TEST_COLLINDEXED_TABLE_INDEXES: list[tuple[str, str | dict[str, str]]] = [
    ("test_collidxtable_idx_si", "set_int"),
    ("test_collidxtable_idx_li", "list_int"),
    ("test_collidxtable_idx_mtie", "map_text_int_e"),
    ("test_collidxtable_idx_mtik", {"map_text_int_k": "$keys"}),
    ("test_collidxtable_idx_mtiv", {"map_text_int_v": "$values"}),
]

TEST_TEXTINDEX_TABLE_NAME = "test_table_textindex"
TEST_TEXTINDEX_TABLE_DEFINITION = CreateTableDefinition(
    columns={
        "id": TableScalarColumnTypeDescriptor(column_type="text"),
        "body": TableScalarColumnTypeDescriptor(column_type="text"),
    },
    primary_key=TablePrimaryKeyDescriptor(
        partition_by=["id"],
        partition_sort={},
    ),
)
TEST_TEXTINDEX_INDEX_NAME = "test_table_textindex_idx"
TEST_TEXTINDEX_INDEX_COLUMN = "body"

TEST_LOGICALFILTERING_TABLE_NAME = "test_table_logicalfiltering"
TEST_LOGICALFILTERING_TABLE_DEFINITION = CreateTableDefinition(
    columns={
        "id": TableScalarColumnTypeDescriptor(column_type="text"),
        "p_boolean": TableScalarColumnTypeDescriptor(column_type="boolean"),
        "p_int_1": TableScalarColumnTypeDescriptor(column_type="int"),
        "p_int_2": TableScalarColumnTypeDescriptor(column_type="int"),
        "p_text": TableScalarColumnTypeDescriptor(column_type="text"),
    },
    primary_key=TablePrimaryKeyDescriptor(
        partition_by=["id"],
        partition_sort={},
    ),
)
TEST_LOGICALFILTERING_TABLE_INDEXES: list[tuple[str, str | dict[str, str]]] = [
    ("test_logflt_idx_p_bool", "p_boolean"),
    ("test_logflt_idx_p_int_1", "p_int_1"),
    ("test_logflt_idx_p_int_2", "p_int_2"),
    ("test_logflt_idx_p_text", "p_text"),
]

VECTORIZE_TEXTS = [
    "The world is the totality of facts, not of things.",
    "The world is determined by the facts, and by these being all the facts.",
    "For the totality of facts determines both what is the case, and also all that is not the case.",
    "The facts in logical space are the world.",
    "The world divides into facts.",
    "Any one can either be the case or not be the case, and everything else remain the same.",
    "What is the case, the fact, is the existence of atomic facts.",
    "An atomic fact is a combination of objects (entities, things).",
    "It is essential to a thing that it can be a constituent part of an atomic fact.",
    "In logic nothing is accidental: if a thing can occur in an atomic fact the possibility of that atomic fact must already be prejudged in the thing.",
    "It would, so to speak, appear as an accident, when to a thing that could exist alone on its own account, subsequently a state of affairs could be made to fit.",
    "If things can occur in atomic facts, this possibility must already lie in them.",
    "(A logical entity cannot be merely possible. Logic treats of every possibility, and all possibilities are its facts.)",
    "Just as we cannot think of spatial objects at all apart from space, or temporal objects apart from time, so we cannot think of any object apart from the possibility of its connection with other things.",
    "If I can think of an object in the context of an atomic fact, I cannot think of it apart from the possibility of this context.",
    "The thing is independent, in so far as it can occur in all possible circumstances, but this form of independence is a form of connection with the atomic fact, a form of dependence. (It is impossible for words to occur in two different ways, alone and in the proposition.)",
    "If I know an object, then I also know all the possibilities of its occurrence in atomic facts.",
    "(Every such possibility must lie in the nature of the object.)",
    "A new possibility cannot subsequently be found.",
    "In order to know an object, I must know not its external but all its internal qualities.",
    "If all objects are given, then thereby are all possible atomic facts also given.",
    "Every thing is, as it were, in a space of possible atomic facts. I can think of this space as empty, but not of the thing without the space.",
    "A spatial object must lie in infinite space. (A point in space is an argument place.)",
    "A speck in a visual field need not be red, but it must have a colour; it has, so to speak, a colour space round it. A tone must have a pitch, the object of the sense of touch a hardness, etc.",
    "Objects contain the possibility of all states of affairs.",
    "The possibility of its occurrence in atomic facts is the form of the object.",
    "The object is simple.",
]


def dict_equal_same_class(
    dct1: dict[Any, Any] | None, dct2: dict[Any, Any] | None
) -> None:
    """A strong equality that checks the class in each value is exactly the same."""
    if dct1 is None:
        assert dct2 is None
    elif dct2 is None:
        assert dct1 is None
    else:
        assert dct1 == dct2
        for k in dct1.keys() | dct2.keys():
            assert dct1 == dct2
            assert dct1[k].__class__ == dct2[k].__class__



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import datetime
import pickle

import pytest

from astrapy.data_types import DataAPIDate


class TestDataAPIDate:
    @pytest.mark.describe("test of date type, errors in parsing from string")
    def test_dataapidate_parse_errors(self) -> None:
        # empty, faulty, misformatted
        with pytest.raises(ValueError):
            DataAPIDate.from_string("")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("boom")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("1-1-1-1")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("1-1")

        # signs for various lengths of the 'year digits' portion
        # valid cases
        DataAPIDate.from_string("0000-10-29")
        DataAPIDate.from_string("-2024-10-29")
        DataAPIDate.from_string("+112024-10-29")
        DataAPIDate.from_string("+000024-10-29")
        DataAPIDate.from_string("-112024-10-29")
        DataAPIDate.from_string("0124-10-29")
        DataAPIDate.from_string("-0124-10-29")
        # incorrect cases for sign of year
        with pytest.raises(ValueError):
            DataAPIDate.from_string("+0000-10-29")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("-0000-10-29")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("+124-10-29")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("+2024-10-29")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("+-2024-10-29")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("-+2024-10-29")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("124-10-29")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("+124-10-29")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("-124-10-29")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("123456-10-29")

        with pytest.raises(ValueError):
            DataAPIDate.from_string("+1999-1-1")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-10-10X")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("X2024-10-10")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("X2024-X10-10")

        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-0-1")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-1-0")

        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-13-20")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-04-31")

        DataAPIDate.from_string("2024-01-31")
        DataAPIDate.from_string("2023-02-28")
        DataAPIDate.from_string("2024-02-29")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2023-02-29")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-02-30")

        DataAPIDate.from_string("2024-03-31")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-03-32")
        DataAPIDate.from_string("2024-04-30")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-04-31")
        DataAPIDate.from_string("2024-05-31")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-05-32")
        DataAPIDate.from_string("2024-06-30")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-06-31")
        DataAPIDate.from_string("2024-07-31")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-07-32")
        DataAPIDate.from_string("2024-08-31")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-08-32")
        DataAPIDate.from_string("2024-09-30")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-09-31")
        DataAPIDate.from_string("2024-10-31")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-10-32")
        DataAPIDate.from_string("2024-11-30")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-11-31")
        DataAPIDate.from_string("2024-12-31")
        with pytest.raises(ValueError):
            DataAPIDate.from_string("2024-12-32")

    @pytest.mark.describe(
        "test of DataAPIData sign-of-year parse and stringify consistency"
    )
    def test_dataapidate_sign_of_year_consistency(self) -> None:
        def parse_and_back_year_str(year_str: str) -> None:
            # just testing that years come out as they get in with numdigits, sign
            dt_string = f"{year_str}-01-12"
            dt = DataAPIDate.from_string(dt_string)
            dt_string_1 = dt.to_string()
            assert dt_string == dt_string_1

        parse_and_back_year_str("2024")
        parse_and_back_year_str("0000")
        parse_and_back_year_str("-2024")
        parse_and_back_year_str("+112024")
        parse_and_back_year_str("-112024")
        parse_and_back_year_str("0124")
        parse_and_back_year_str("-0124")
        parse_and_back_year_str("-123456")
        parse_and_back_year_str("1972")

    @pytest.mark.describe("test of date type, lifecycle")
    def test_dataapidate_lifecycle(self) -> None:
        dfull = DataAPIDate.from_string("1995-02-09")
        dfull_exp = DataAPIDate(year=1995, month=2, day=9)
        assert dfull == dfull_exp
        assert dfull == DataAPIDate.from_string("1995-2-9")
        assert dfull == DataAPIDate.from_string("+00001995-0002-0009")

        py_dfull = datetime.date(1995, 2, 9)
        assert DataAPIDate.from_date(py_dfull) == dfull
        assert dfull.to_date() == py_dfull

        DataAPIDate.from_string("-55000-03-30")
        DataAPIDate.from_string("+1234567-12-31")

        repr(dfull)
        str(dfull)
        dfull.to_string()

        end_pleistocene = DataAPIDate.from_string("-11700-12-13")
        end_pleistocene.to_string()
        with pytest.raises(ValueError):
            end_pleistocene.to_date()

        far_future = DataAPIDate.from_string("+252525-01-24")
        far_future.to_string()
        with pytest.raises(ValueError):
            far_future.to_date()

        assert far_future > end_pleistocene
        assert far_future >= end_pleistocene
        assert end_pleistocene < far_future
        assert end_pleistocene <= far_future

        modernity = datetime.date(2024, 10, 28)
        assert far_future > modernity
        assert far_future >= modernity
        assert modernity < far_future
        assert modernity <= far_future

        assert end_pleistocene < modernity
        assert end_pleistocene <= modernity
        assert modernity > end_pleistocene
        assert modernity >= end_pleistocene

    @pytest.mark.describe("test pickling of DataAPIDate")
    def test_dataapidate_pickle(self) -> None:
        the_date = DataAPIDate(2024, 3, 14)
        assert pickle.loads(pickle.dumps(the_date)) == the_date



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from datetime import date, datetime, timezone
from typing import Any

import pytest

from astrapy.constants import DefaultDocumentType, ReturnDocument, SortMode
from astrapy.cursors import AsyncCollectionFindCursor, CursorState
from astrapy.data_types import (
    DataAPIDate,
    DataAPIMap,
    DataAPITimestamp,
    DataAPIVector,
)
from astrapy.exceptions import (
    CollectionInsertManyException,
    DataAPIResponseException,
    TooManyDocumentsToCountException,
)
from astrapy.ids import UUID, ObjectId
from astrapy.results import CollectionDeleteResult, CollectionInsertOneResult
from astrapy.utils.api_options import APIOptions, SerdesOptions

from ..conftest import DefaultAsyncCollection


async def _alist(
    acursor: AsyncCollectionFindCursor[DefaultDocumentType, Any],
) -> list[Any]:
    return [doc async for doc in acursor]


class TestCollectionDMLAsync:
    @pytest.mark.describe("test of collection count_documents, async")
    async def test_collection_count_documents_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
            == 0
        )
        await async_empty_collection.insert_one({"doc": 1, "group": "A"})
        await async_empty_collection.insert_one({"doc": 2, "group": "B"})
        await async_empty_collection.insert_one({"doc": 3, "group": "A"})
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
            == 3
        )
        assert (
            await async_empty_collection.count_documents(
                filter={"group": "A"}, upper_bound=100
            )
            == 2
        )

    @pytest.mark.describe("test of collection estimated_document_count, async")
    async def test_collection_estimated_document_count_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        await async_empty_collection.insert_one({"doc": 1, "group": "A"})
        await async_empty_collection.insert_one({"doc": 2, "group": "B"})
        await async_empty_collection.insert_one({"doc": 3, "group": "A"})
        count = await async_empty_collection.estimated_document_count()
        # it's _estimated_, no precise expectation with such short sizes/times
        assert isinstance(count, int)
        assert count >= 0

    @pytest.mark.describe("test of overflowing collection count_documents, async")
    async def test_collection_overflowing_count_documents_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        await async_empty_collection.insert_many([{"a": i} for i in range(900)])
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=950)
            == 900
        )
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=2000)
            == 900
        )
        with pytest.raises(TooManyDocumentsToCountException):
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
        await async_empty_collection.insert_many([{"b": i} for i in range(200)])
        with pytest.raises(TooManyDocumentsToCountException):
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
        with pytest.raises(TooManyDocumentsToCountException):
            await async_empty_collection.count_documents(filter={}, upper_bound=2000)

    @pytest.mark.describe("test of collection insert_one, async")
    async def test_collection_insert_one_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        io_result1 = await async_empty_collection.insert_one({"doc": 1, "group": "A"})
        assert isinstance(io_result1, CollectionInsertOneResult)
        io_result2 = await async_empty_collection.insert_one(
            {"_id": "xxx", "doc": 2, "group": "B"}
        )
        assert io_result2.inserted_id == "xxx"
        assert (
            await async_empty_collection.count_documents(
                filter={"group": "A"}, upper_bound=100
            )
            == 1
        )

    @pytest.mark.describe("test of collection insert_one with vector, async")
    async def test_collection_insert_one_vector_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        await async_empty_collection.insert_one({"tag": "v1", "$vector": [-1, -2]})
        retrieved1 = await async_empty_collection.find_one(
            {"tag": "v1"}, projection={"*": 1}
        )
        assert retrieved1 is not None
        assert retrieved1["$vector"] == DataAPIVector([-1, -2])

        await async_empty_collection.insert_one({"tag": "v2", "$vector": [-3, -4]})
        retrieved2 = await async_empty_collection.find_one(
            {"tag": "v2"}, projection={"*": 1}
        )
        assert retrieved2 is not None
        assert retrieved2["$vector"] == DataAPIVector([-3, -4])

        await async_empty_collection.insert_one({"tag": "v_null", "$vector": None})
        retrieved3 = await async_empty_collection.find_one(
            {"tag": "v_null"}, projection={"*": 1}
        )
        assert retrieved3 is not None
        assert retrieved3["$vector"] is None

        unroll_options = APIOptions(
            serdes_options=SerdesOptions(unroll_iterables_to_lists=True),
        )
        unroll_acoll = async_empty_collection.with_options(api_options=unroll_options)
        await unroll_acoll.insert_one({"tag": "v_null_u", "$vector": None})
        retrieved4 = await unroll_acoll.find_one(
            {"tag": "v_null_u"}, projection={"*": 1}
        )
        assert retrieved4 is not None
        assert retrieved4["$vector"] is None

    @pytest.mark.describe("test of collection vector insertion options, async")
    async def test_collection_vector_insertion_options_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acollection_Yb_Yc = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(
                    binary_encode_vectors=True,
                    unroll_iterables_to_lists=True,
                ),
            ),
        )
        acollection_Nb_Yc = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(
                    binary_encode_vectors=False,
                    unroll_iterables_to_lists=True,
                ),
            ),
        )
        acollection_Yb_Nc = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(
                    binary_encode_vectors=True,
                    unroll_iterables_to_lists=False,
                ),
            ),
        )
        acollection_Nb_Nc = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(
                    binary_encode_vectors=False,
                    unroll_iterables_to_lists=False,
                ),
            ),
        )

        # writes w.r.t. options
        await acollection_Yb_Yc.insert_one(
            {"_id": "Yb_Yc_()", "$vector": (i for i in [0.1, 0.2])}
        )
        await acollection_Yb_Yc.insert_one({"_id": "Yb_Yc_[]", "$vector": [0.3, 0.4]})
        await acollection_Yb_Yc.insert_one(
            {"_id": "Yb_Yc_DV", "$vector": DataAPIVector([0.5, 0.6])}
        )

        await acollection_Nb_Yc.insert_one(
            {"_id": "Nb_Yc_()", "$vector": (i for i in [0.1, 0.2])}
        )
        await acollection_Nb_Yc.insert_one({"_id": "Nb_Yc_[]", "$vector": [0.3, 0.4]})
        await acollection_Nb_Yc.insert_one(
            {"_id": "Nb_Yc_DV", "$vector": DataAPIVector([0.5, 0.6])}
        )

        with pytest.raises(TypeError):
            await acollection_Yb_Nc.insert_one(
                {"_id": "Yb_Nc_()", "$vector": (i for i in [0.1, 0.2])}
            )
        await acollection_Yb_Nc.insert_one({"_id": "Yb_Nc_[]", "$vector": [0.3, 0.4]})
        await acollection_Yb_Nc.insert_one(
            {"_id": "Yb_Nc_DV", "$vector": DataAPIVector([0.5, 0.6])}
        )

        with pytest.raises(TypeError):
            await acollection_Nb_Nc.insert_one(
                {"_id": "Nb_Nc_()", "$vector": (i for i in [0.1, 0.2])}
            )
        await acollection_Nb_Nc.insert_one({"_id": "Nb_Nc_[]", "$vector": [0.3, 0.4]})
        await acollection_Nb_Nc.insert_one(
            {"_id": "Nb_Nc_DV", "$vector": DataAPIVector([0.5, 0.6])}
        )

        # check how the documents are stored
        expect_binaries = {
            "Yb_Yc_()": True,
            "Yb_Yc_[]": True,
            "Yb_Yc_DV": True,
            "Nb_Yc_()": False,
            "Nb_Yc_[]": False,
            "Nb_Yc_DV": False,
            #
            "Yb_Nc_[]": True,
            "Yb_Nc_DV": True,
            #
            "Nb_Nc_[]": False,
            "Nb_Nc_DV": False,
        }

        raw_find_response = await async_empty_collection.command(
            body={"find": {"projection": {"_id": True, "$vector": True}}},
        )
        raw_docs = raw_find_response["data"]["documents"]
        for raw_doc in raw_docs:
            expect_binary = expect_binaries[raw_doc["_id"]]
            has_binary = "$binary" in raw_doc["$vector"]
            assert expect_binary is has_binary

        acollection_Ycc = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=True),
            ),
        )
        acollection_Ncc = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=False),
            ),
        )
        docs_Ycc = [
            doc async for doc in acollection_Ycc.find(projection={"$vector": True})
        ]
        docs_Ncc = [
            doc async for doc in acollection_Ncc.find(projection={"$vector": True})
        ]
        assert all(isinstance(doc["$vector"], DataAPIVector) for doc in docs_Ycc)
        assert all(not isinstance(doc["$vector"], DataAPIVector) for doc in docs_Ncc)

    @pytest.mark.describe("test of collection delete_one, async")
    async def test_collection_delete_one_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        await async_empty_collection.insert_one({"doc": 1, "group": "A"})
        await async_empty_collection.insert_one({"doc": 2, "group": "B"})
        await async_empty_collection.insert_one({"doc": 3, "group": "A"})
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
            == 3
        )
        do_result1 = await async_empty_collection.delete_one({"group": "A"})
        assert isinstance(do_result1, CollectionDeleteResult)
        assert do_result1.deleted_count == 1
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
            == 2
        )

        # test of sort
        await async_empty_collection.insert_many(
            [{"ts": 1, "seq": i} for i in [2, 0, 1]]
        )
        await async_empty_collection.delete_one({"ts": 1}, sort={"seq": 1})
        assert set(await async_empty_collection.distinct("seq", filter={"ts": 1})) == {
            1,
            2,
        }

    @pytest.mark.describe("test of collection delete_many, async")
    async def test_collection_delete_many_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        await async_empty_collection.insert_one({"doc": 1, "group": "A"})
        await async_empty_collection.insert_one({"doc": 2, "group": "B"})
        await async_empty_collection.insert_one({"doc": 3, "group": "A"})
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
            == 3
        )
        do_result1 = await async_empty_collection.delete_many({"group": "A"})
        assert isinstance(do_result1, CollectionDeleteResult)
        assert do_result1.deleted_count == 2
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
            == 1
        )

        await async_empty_collection.delete_many({})
        await async_empty_collection.insert_many([{"a": 1} for _ in range(50)])
        do_result2 = await async_empty_collection.delete_many({"a": 1})
        assert do_result2.deleted_count == 50
        assert await async_empty_collection.count_documents({}, upper_bound=100) == 0

        await async_empty_collection.delete_many({})
        await async_empty_collection.insert_many([{"a": 1} for _ in range(50)])
        do_result2 = await async_empty_collection.delete_many({"a": 1})
        assert do_result2.deleted_count == 50
        assert await async_empty_collection.count_documents({}, upper_bound=100) == 0

    @pytest.mark.describe("test of collection chunk-requiring delete_many, async")
    async def test_collection_chunked_delete_many_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        await async_empty_collection.insert_many(
            [{"doc": i, "group": "A"} for i in range(50)]
        )
        await async_empty_collection.insert_many(
            [{"doc": i, "group": "B"} for i in range(10)]
        )
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
        ) == 60
        do_result1 = await async_empty_collection.delete_many({"group": "A"})
        assert isinstance(do_result1, CollectionDeleteResult)
        assert do_result1.deleted_count == 50
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
        ) == 10

    @pytest.mark.describe("test of collection find, async")
    async def test_collection_find_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        await async_empty_collection.insert_many([{"seq": i} for i in range(30)])
        Nski = 1
        Nlim = 28
        Nsor = {"seq": SortMode.DESCENDING}
        Nfil = {"seq": {"$exists": True}}

        # case 0000 of find-pattern matrix
        assert (
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=None, limit=None, sort=None, filter=None
                    )
                )
            )
            == 30
        )

        # case 0001
        assert (
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=None, limit=None, sort=None, filter=Nfil
                    )
                )
            )
            == 30
        )

        # case 0010
        assert (
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=None, limit=None, sort=Nsor, filter=None
                    )
                )
            )
            == 20
        )  # NONPAGINATED

        # case 0011
        assert (
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=None, limit=None, sort=Nsor, filter=Nfil
                    )
                )
            )
            == 20
        )  # NONPAGINATED

        # case 0100
        assert (
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=None, limit=Nlim, sort=None, filter=None
                    )
                )
            )
            == 28
        )

        # case 0101
        assert (
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=None, limit=Nlim, sort=None, filter=Nfil
                    )
                )
            )
            == 28
        )

        # case 0110
        assert (
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=None, limit=Nlim, sort=Nsor, filter=None
                    )
                )
            )
            == 20
        )  # NONPAGINATED

        # case 0111
        assert (
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=None, limit=Nlim, sort=Nsor, filter=Nfil
                    )
                )
            )
            == 20
        )  # NONPAGINATED

        # case 1000
        with pytest.raises(DataAPIResponseException):
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=Nski, limit=None, sort=None, filter=None
                    )
                )
            )

        # case 1001
        with pytest.raises(DataAPIResponseException):
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=Nski, limit=None, sort=None, filter=Nfil
                    )
                )
            )

        # case 1010
        assert (
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=Nski, limit=None, sort=Nsor, filter=None
                    )
                )
            )
            == 20
        )  # NONPAGINATED

        # case 1011
        assert (
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=Nski, limit=None, sort=Nsor, filter=Nfil
                    )
                )
            )
            == 20
        )  # NONPAGINATED

        # case 1100
        with pytest.raises(DataAPIResponseException):
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=Nski, limit=Nlim, sort=None, filter=None
                    )
                )
            )

        # case 1101
        with pytest.raises(DataAPIResponseException):
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=Nski, limit=Nlim, sort=None, filter=Nfil
                    )
                )
            )

        # case 1110
        assert (
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=Nski, limit=Nlim, sort=Nsor, filter=None
                    )
                )
            )
            == 20
        )  # NONPAGINATED

        # case 1111
        assert (
            len(
                await _alist(
                    async_empty_collection.find(
                        skip=Nski, limit=Nlim, sort=Nsor, filter=Nfil
                    )
                )
            )
            == 20
        )  # NONPAGINATED

    @pytest.mark.describe("test of cursors from collection.find, async")
    async def test_collection_cursors_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        """
        Functionalities of cursors from find, other than the various
        combinations of skip/limit/sort/filter specified above.
        """
        await async_empty_collection.insert_many(
            [{"seq": i, "ternary": (i % 3)} for i in range(10)]
        )

        # projection
        cursor0 = async_empty_collection.find(projection={"ternary": False})
        assert cursor0.consumed == 0
        document0 = await cursor0.__anext__()
        assert cursor0.consumed == 1
        assert "ternary" not in document0
        cursor0b = async_empty_collection.find(projection={"ternary": True})
        document0b = await cursor0b.__anext__()
        assert "ternary" in document0b

        assert cursor0b.data_source == async_empty_collection

        # rewinding, slicing and retrieved
        cursor1 = async_empty_collection.find(sort={"seq": 1})
        await cursor1.__anext__()
        await cursor1.__anext__()
        items1 = (await _alist(cursor1))[:2]  # noqa: F841
        cursor1.rewind()
        assert await _alist(cursor1) == await _alist(
            async_empty_collection.find(sort={"seq": 1})
        )
        cursor1.rewind()

        # address, cursor_id, collection
        assert isinstance(cursor1.cursor_id, int)
        assert cursor1.data_source == async_empty_collection

        # clone
        cursor2 = async_empty_collection.find()
        assert cursor2.state != CursorState.CLOSED
        for _ in range(8):
            await cursor2.__anext__()
        assert cursor2.state != CursorState.CLOSED  # type: ignore[comparison-overlap]
        cursor3 = cursor2.clone()
        assert len(await _alist(cursor2)) == 2
        assert len(await _alist(cursor3)) == 10
        assert cursor2.state == CursorState.CLOSED  # type: ignore[comparison-overlap]

        # close
        cursor4 = async_empty_collection.find()
        for _ in range(8):
            await cursor4.__anext__()
        cursor4.close()
        assert cursor4.state == CursorState.CLOSED
        with pytest.raises(StopAsyncIteration):
            await cursor4.__anext__()

        # distinct from collections
        assert set(await async_empty_collection.distinct("ternary")) == {0, 1, 2}
        assert set(await async_empty_collection.distinct("nonfield")) == set()

    @pytest.mark.describe(
        "test of collective and mappings on cursors from collection.find, async"
    )
    async def test_collection_cursors_collective_maps_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        """
        Functionalities of cursors from find: map, tolist, foreach.
        """
        await async_empty_collection.insert_many(
            [{"seq": i, "ternary": (i % 3)} for i in range(75)]
        )

        cur0 = async_empty_collection.find(filter={"seq": {"$gte": 10}})
        assert len(await _alist(cur0)) == 65

        # map
        def mapper(doc: DefaultDocumentType) -> tuple[int, str]:
            return (doc["seq"], f"T{doc['ternary']}")

        cur1_map = async_empty_collection.find(filter={"seq": {"$gte": 10}}).map(mapper)
        items1_map = await _alist(cur1_map)
        assert len(items1_map) == 65
        assert (20, "T2") in items1_map
        assert all(isinstance(tup[0], int) for tup in items1_map)
        assert all(isinstance(tup[1], str) for tup in items1_map)

        # map composition

        def mapper_2(tup: tuple[int, str]) -> str:
            return f"{tup[1]}/{tup[0]}"

        cur2_maps = (
            async_empty_collection.find(filter={"seq": {"$gte": 10}})
            .map(mapper)
            .map(mapper_2)
        )
        items2_maps = await _alist(cur2_maps)
        assert len(items2_maps) == 65
        assert "T2/20" in items2_maps
        assert all(isinstance(itm, str) for itm in items2_maps)

        # clone (rewinding)
        cloned_2 = cur2_maps.clone()
        from_cl = await cloned_2.__anext__()
        assert isinstance(from_cl, str)
        assert "/" in from_cl

        # for each
        accum: list[int] = []

        def ingest(doc: DefaultDocumentType, acc: list[int] = accum) -> None:
            acc += [doc["ternary"]]

        await async_empty_collection.find(filter={"seq": {"$gte": 10}}).for_each(ingest)
        assert len(accum) == 65
        assert set(accum) == {0, 1, 2}

        # to list
        cur3_tl = async_empty_collection.find(
            filter={"seq": {"$gte": 10}},
            projection={"_id": False},
        )
        items3_tl = await cur3_tl.to_list()
        assert len(items3_tl) == 65
        assert {"seq": 50, "ternary": 2} in items3_tl

    @pytest.mark.describe("test of distinct with non-hashable items, async")
    async def test_collection_distinct_nonhashable_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection
        the_datimestamp = DataAPITimestamp.from_string("2000-01-01T12:00:00.000Z")
        documents: list[dict[str, Any]] = [
            {},
            {"f": 1},
            {"f": "a"},
            {"f": {"subf": 99}},
            {"f": {"subf": 99, "another": {"subsubf": [True, False]}}},
            {"f": [10, 11]},
            {"f": [11, 10]},
            {"f": [10]},
            {"f": the_datimestamp},
            {"f": None},
        ]
        await acol.insert_many(documents * 2)

        d_items = await acol.distinct("f")
        assert len(d_items) == 8
        for doc in documents:
            if "f" in doc:
                if isinstance(doc["f"], list):
                    for item in doc["f"]:
                        assert item in d_items
                else:
                    assert doc["f"] in d_items

        d_items_noncustom = await acol.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=False)
            ),
        ).distinct("f")
        assert len(d_items_noncustom) == 8
        for doc in documents:
            if "f" in doc:
                if isinstance(doc["f"], list):
                    for item in doc["f"]:
                        assert item in d_items_noncustom
                elif isinstance(doc["f"], DataAPITimestamp):
                    assert doc["f"].to_datetime(tz=timezone.utc) in d_items_noncustom
                else:
                    assert doc["f"] in d_items_noncustom

    @pytest.mark.describe("test of distinct with key as list, async")
    async def test_collection_distinct_key_as_list_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection
        the_datimestamp = DataAPITimestamp.from_string("2000-01-01T12:00:00.000Z")
        documents: list[dict[str, Any]] = [
            {},
            {"f": 1},
            {"f": "a"},
            {"f": {"subf": 99}},
            {"f": {"subf": 99, "another": {"subsubf": [True, False]}}},
            {"f": [10, 11]},
            {"f": [11, 10]},
            {"f": [10]},
            {"f": the_datimestamp},
            {"f": None},
        ]
        await acol.insert_many(documents * 2)

        assert await acol.distinct("f") == await acol.distinct(["f"])

        await acol.insert_one({"x": [{"y": "Y", "0": "ZERO"}]})

        await acol.delete_many({})
        await acol.insert_one({"x": [{"y": "Y", "0": "ZERO"}]})

        assert await acol.distinct(["x", "y"]) == ["Y"]
        # these expose int-vs-str subtleties (listindex/mapkey issues)
        assert await acol.distinct(["x", "0"]) == ["ZERO"]
        assert await acol.distinct(["x", 0]) == [{"y": "Y", "0": "ZERO"}]
        assert await acol.distinct(["x", "0", "y"]) == []
        assert await acol.distinct(["x", 0, "y"]) == ["Y"]
        assert await acol.distinct(["x", "0", "0"]) == []
        assert await acol.distinct(["x", "0", 0]) == []
        assert await acol.distinct(["x", 0, "0"]) == ["ZERO"]
        assert await acol.distinct(["x", 0, 0]) == []

        with pytest.raises(ValueError):
            await async_empty_collection.distinct(["root", "1", "", "subf"])
        with pytest.raises(ValueError):
            await async_empty_collection.distinct(["root", "", "1", "subf"])
        with pytest.raises(ValueError):
            await async_empty_collection.distinct(["root", "", "subf", "subsubf"])
        with pytest.raises(ValueError):
            await async_empty_collection.distinct(["root", "subf", "", "subsubf"])

    @pytest.mark.describe("test of usage of projection in distinct, async")
    async def test_collection_projections_distinct_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection
        await acol.insert_one({"x": [{"y": "Y", "0": "ZERO"}]})

        assert await acol.distinct("x.y") == ["Y"]
        # the one below shows that if index-in-list, then browse-whole-list is off
        assert await acol.distinct("x.0") == [{"y": "Y", "0": "ZERO"}]
        assert await acol.distinct("x.0.y") == ["Y"]
        assert await acol.distinct("x.0.0") == ["ZERO"]

    @pytest.mark.describe("test of unacceptable paths for distinct, async")
    async def test_collection_wrong_paths_distinct_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        with pytest.raises(ValueError):
            await async_empty_collection.distinct("root.1..subf")
        with pytest.raises(ValueError):
            await async_empty_collection.distinct("root..1.subf")
        with pytest.raises(ValueError):
            await async_empty_collection.distinct("root..subf.subsubf")
        with pytest.raises(ValueError):
            await async_empty_collection.distinct("root.subf..subsubf")

    @pytest.mark.describe("test of collection find with vectors, async")
    async def test_collection_find_find_one_vectors_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        q_vector = [3, 3]
        await async_empty_collection.insert_many(
            [
                {"tag": "A", "$vector": [4, 5]},
                {"tag": "B", "$vector": [3, 4]},
                {"tag": "C", "$vector": [3, 2]},
                {"tag": "D", "$vector": [4, 1]},
                {"tag": "E", "$vector": [2, 5]},
            ]
        )

        hits = [
            hit
            async for hit in async_empty_collection.find(
                {},
                sort={"$vector": q_vector},
                projection=["tag"],
                limit=3,
            )
        ]
        assert [hit["tag"] for hit in hits] == ["A", "B", "C"]

        with pytest.raises(DataAPIResponseException):
            [
                itm
                async for itm in async_empty_collection.find(
                    {},
                    projection=["tag"],
                    sort={"$vector": q_vector, "tag": SortMode.DESCENDING},
                    limit=3,
                )
            ]

        top_doc = await async_empty_collection.find_one({}, sort={"$vector": [1, 0]})
        assert top_doc is not None
        assert top_doc["tag"] == "D"

        fdoc_no_s = await async_empty_collection.find(
            {}, sort={"$vector": [1, 1]}, include_similarity=False
        ).__anext__()
        fdoc_wi_s = await async_empty_collection.find(
            {}, sort={"$vector": [1, 1]}, include_similarity=True
        ).__anext__()
        assert fdoc_no_s is not None
        assert fdoc_wi_s is not None
        assert "$similarity" not in fdoc_no_s
        assert "$similarity" in fdoc_wi_s
        assert fdoc_wi_s["$similarity"] > 0.0

        f1doc_no_s = await async_empty_collection.find_one(
            {}, sort={"$vector": [1, 1]}, include_similarity=False
        )
        f1doc_wi_s = await async_empty_collection.find_one(
            {}, sort={"$vector": [1, 1]}, include_similarity=True
        )
        assert f1doc_no_s is not None
        assert f1doc_wi_s is not None
        assert "$similarity" not in f1doc_no_s
        assert "$similarity" in f1doc_wi_s
        assert f1doc_wi_s["$similarity"] > 0.0

    @pytest.mark.describe("test of include_sort_vector in collection find, async")
    async def test_collection_include_sort_vector_find_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        q_vector = DataAPIVector([10, 9])

        # with empty collection
        for include_sv in [False, True]:
            for sort_cl_label in ["reg", "vec"]:
                sort_cl_e: dict[str, Any] = (
                    {} if sort_cl_label == "reg" else {"$vector": q_vector}
                )
                vec_expected = include_sv and sort_cl_label == "vec"
                # pristine iterator
                this_ite_1 = async_empty_collection.find(
                    {}, sort=sort_cl_e, include_sort_vector=include_sv
                )
                if vec_expected:
                    assert (await this_ite_1.get_sort_vector()) == q_vector
                else:
                    assert (await this_ite_1.get_sort_vector()) is None
                # after exhaustion with empty
                all_items_1 = await _alist(this_ite_1)
                assert all_items_1 == []
                if vec_expected:
                    assert (await this_ite_1.get_sort_vector()) == q_vector
                else:
                    assert (await this_ite_1.get_sort_vector()) is None
                # directly exhausted before calling get_sort_vector
                this_ite_2 = async_empty_collection.find(
                    {}, sort=sort_cl_e, include_sort_vector=include_sv
                )
                all_items_2 = await _alist(this_ite_2)
                assert all_items_2 == []
                if vec_expected:
                    assert (await this_ite_1.get_sort_vector()) == q_vector
                else:
                    assert (await this_ite_1.get_sort_vector()) is None
        await async_empty_collection.insert_many(
            [{"seq": i, "$vector": [i, i + 1]} for i in range(10)]
        )
        # with non-empty collection
        for include_sv in [False, True]:
            for sort_cl_label in ["reg", "vec"]:
                sort_cl_f: dict[str, Any] = (
                    {} if sort_cl_label == "reg" else {"$vector": q_vector}
                )
                vec_expected = include_sv and sort_cl_label == "vec"
                # pristine iterator
                this_ite_1 = async_empty_collection.find(
                    {}, sort=sort_cl_f, include_sort_vector=include_sv
                )
                if vec_expected:
                    assert (await this_ite_1.get_sort_vector()) == q_vector
                else:
                    assert (await this_ite_1.get_sort_vector()) is None
                # after consuming one item
                first_seqs = [
                    doc["seq"]
                    for doc in [
                        await this_ite_1.__anext__(),
                        await this_ite_1.__anext__(),
                    ]
                ]
                if vec_expected:
                    assert (await this_ite_1.get_sort_vector()) == q_vector
                else:
                    assert (await this_ite_1.get_sort_vector()) is None
                # after exhaustion with the rest
                last_seqs = [doc["seq"] async for doc in this_ite_1]
                assert len(set(last_seqs + first_seqs)) == 10
                assert len(last_seqs + first_seqs) == 10
                if vec_expected:
                    assert (await this_ite_1.get_sort_vector()) == q_vector
                else:
                    assert (await this_ite_1.get_sort_vector()) is None
                # directly exhausted before calling get_sort_vector
                this_ite_2 = async_empty_collection.find(
                    {}, sort=sort_cl_f, include_sort_vector=include_sv
                )
                await _alist(this_ite_2)
                if vec_expected:
                    assert (await this_ite_1.get_sort_vector()) == q_vector
                else:
                    assert (await this_ite_1.get_sort_vector()) is None

    @pytest.mark.describe("test of include_sort_vector with serdes options, async")
    async def test_collection_include_sort_vector_serdes_options_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol_v0_d0 = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(
                    custom_datatypes_in_reading=False,
                    use_decimals_in_collections=False,
                ),
            ),
        )
        acol_v0_d1 = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(
                    custom_datatypes_in_reading=False,
                    use_decimals_in_collections=True,
                ),
            ),
        )
        acol_v1_d0 = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(
                    custom_datatypes_in_reading=True,
                    use_decimals_in_collections=False,
                ),
            ),
        )
        acol_v1_d1 = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(
                    custom_datatypes_in_reading=True,
                    use_decimals_in_collections=True,
                ),
            ),
        )
        cur0_v0_d0 = acol_v0_d0.find(sort={"$vector": [1, 2]})
        cur0_v0_d1 = acol_v0_d1.find(sort={"$vector": [1, 2]})
        cur0_v1_d0 = acol_v1_d0.find(sort={"$vector": [1, 2]})
        cur0_v1_d1 = acol_v1_d1.find(sort={"$vector": [1, 2]})

        assert await cur0_v0_d0.get_sort_vector() is None
        assert await cur0_v0_d1.get_sort_vector() is None
        assert await cur0_v1_d0.get_sort_vector() is None
        assert await cur0_v1_d1.get_sort_vector() is None

        cur1_v0_d0 = acol_v0_d0.find(sort={"$vector": [1, 2]}, include_sort_vector=True)
        cur1_v0_d1 = acol_v0_d1.find(sort={"$vector": [1, 2]}, include_sort_vector=True)
        cur1_v1_d0 = acol_v1_d0.find(sort={"$vector": [1, 2]}, include_sort_vector=True)
        cur1_v1_d1 = acol_v1_d1.find(sort={"$vector": [1, 2]}, include_sort_vector=True)

        assert await cur1_v0_d0.get_sort_vector() == [1, 2]
        assert await cur1_v0_d1.get_sort_vector() == [1, 2]
        assert await cur1_v1_d0.get_sort_vector() == DataAPIVector([1, 2])
        assert await cur1_v1_d1.get_sort_vector() == DataAPIVector([1, 2])

    @pytest.mark.describe("test of projections in collection find with vectors, async")
    async def test_collection_find_projections_vectors_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        await async_empty_collection.insert_one(
            {
                "$vector": [1, 2],
                "otherfield": "OF",
                "anotherfield": "AF",
                "text": "T",
            }
        )
        req_projections = [
            None,
            {},
            {"text": True},
            {"$vector": True},
            {"text": True, "$vector": True},
        ]
        exp_fieldsets = [
            {"$vector", "_id", "otherfield", "anotherfield", "text"},
            {"$vector", "_id", "otherfield", "anotherfield", "text"},
            {"_id", "text"},
            {
                "$vector",
                "_id",
                "otherfield",
                "anotherfield",
                "text",
            },
            {"$vector", "_id", "text"},
        ]
        for include_similarity in [True, False]:
            for req_projection, exp_fields0 in zip(req_projections, exp_fieldsets):
                vdocs = [
                    doc
                    async for doc in async_empty_collection.find(
                        sort={"$vector": [11, 21]},
                        limit=1,
                        projection=req_projection,
                        include_similarity=include_similarity,
                    )
                ]
                if include_similarity:
                    exp_fields = exp_fields0 | {"$similarity"}
                else:
                    exp_fields = exp_fields0
                # this test should not concern whether $vector is found or not
                # (abiding by the '$vector may or may not be returned' tenet)
                vkeys_novec = set(vdocs[0].keys()) - {"$vector"}
                expkeys_novec = exp_fields - {"$vector"}
                assert vkeys_novec == expkeys_novec
                # but in some cases $vector must be there:
                if "$vector" in (req_projection or set()):
                    assert "$vector" in vdocs[0]

    @pytest.mark.describe("test of collection insert_many with empty list, async")
    async def test_collection_insert_many_empty_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection
        await acol.insert_many([], ordered=True)
        await acol.insert_many([], ordered=False, concurrency=1)
        await acol.insert_many([], ordered=False, concurrency=5)

    @pytest.mark.describe("test of collection insert_many, async")
    async def test_collection_insert_many_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection

        ins_result1 = await acol.insert_many([{"_id": "a"}, {"_id": "b"}], ordered=True)
        assert set(ins_result1.inserted_ids) == {"a", "b"}
        assert {doc["_id"] async for doc in acol.find()} == {"a", "b"}

        with pytest.raises(CollectionInsertManyException):
            await acol.insert_many([{"_id": "a"}, {"_id": "c"}], ordered=True)
        assert {doc["_id"] async for doc in acol.find()} == {"a", "b"}

        with pytest.raises(CollectionInsertManyException):
            await acol.insert_many(
                [{"_id": "c"}, {"_id": "a"}, {"_id": "d"}], ordered=True
            )
        assert {doc["_id"] async for doc in acol.find()} == {"a", "b", "c"}

        with pytest.raises(CollectionInsertManyException):
            await acol.insert_many(
                [{"_id": "c"}, {"_id": "d"}, {"_id": "e"}],
                ordered=False,
            )
        assert {doc["_id"] async for doc in acol.find()} == {"a", "b", "c", "d", "e"}

    @pytest.mark.describe("test of collection insert_many with vectors, async")
    async def test_collection_insert_many_vectors_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection
        await acol.insert_many(
            [
                {"t": 0, "$vector": [0, 1]},
                {"t": 1, "$vector": [1, 0]},
                {"t": 4, "$vector": [0, 3]},
                {"t": 5, "$vector": [3, 0]},
            ]
        )

        vectors = [doc["$vector"] async for doc in acol.find({}, projection={"*": 1})]
        assert all(len(vec) == 2 for vec in vectors)

    @pytest.mark.describe("test of collection insert_many, failures, async")
    async def test_collection_insert_many_failures_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        # The main goal here is to keep the switch to returnDocumentResponses in check.
        N = 110
        ins_res0 = await async_empty_collection.insert_many(
            [{"_id": i} for i in range(N)],
            concurrency=1,
        )
        assert set(ins_res0.inserted_ids) == set(range(N))

        ins_res1 = await async_empty_collection.insert_many(
            [{"_id": N + i} for i in range(N)],
            concurrency=20,
        )
        assert set(ins_res1.inserted_ids) == {N + i for i in range(N)}

        # unordered insertion [good, bad]
        err2: CollectionInsertManyException | None = None
        try:
            await async_empty_collection.insert_many([{"_id": 2 * N}, {"_id": 0}])
        except CollectionInsertManyException as e:
            err2 = e
        assert err2 is not None
        assert len(err2.exceptions) == 1
        assert isinstance(err2.exceptions[0], DataAPIResponseException)
        assert len(err2.exceptions[0].error_descriptors) == 1
        assert err2.inserted_ids == [2 * N]

        # ordered insertion [good, bad, good_skipped]
        err3: CollectionInsertManyException | None = None
        try:
            await async_empty_collection.insert_many(
                [{"_id": 2 * N + 1}, {"_id": 0}, {"_id": 2 * N + 2}],
                ordered=True,
            )
        except CollectionInsertManyException as e:
            err3 = e
        assert err3 is not None
        assert len(err3.exceptions) == 1
        assert isinstance(err3.exceptions[0], DataAPIResponseException)
        assert len(err3.exceptions[0].error_descriptors) == 1
        assert err3.inserted_ids == [2 * N + 1]

    @pytest.mark.describe("test of collection find_one, async")
    async def test_collection_find_one_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        col = async_empty_collection
        await col.insert_many(
            [
                {"_id": "?", "seq": 0, "kind": "punctuation"},
                {"_id": "a", "seq": 1, "kind": "letter"},
                {"_id": "b", "seq": 2, "kind": "letter"},
            ]
        )

        fo1 = await col.find_one({"kind": "frog"})
        assert fo1 is None

        Nsor = {"seq": 1}
        Nfil = {"kind": "letter"}

        # case 00 of find-pattern matrix
        doc00 = await col.find_one(sort=None, filter=None)
        assert doc00 is not None
        assert doc00["seq"] in {0, 1, 2}

        # case 01
        doc01 = await col.find_one(sort=None, filter=Nfil)
        assert doc01 is not None
        assert doc01["seq"] in {1, 2}

        # case 10
        doc10 = await col.find_one(sort=Nsor, filter=None)
        assert doc10 is not None
        assert doc10["seq"] == 0

        # case 11
        doc11 = await col.find_one(sort=Nsor, filter=Nfil)
        assert doc11 is not None
        assert doc11["seq"] == 1

        # projection
        doc_full = await col.find_one(sort=Nsor, filter=Nfil)
        doc_proj = await col.find_one(sort=Nsor, filter=Nfil, projection={"kind": True})
        assert doc_proj == {"_id": "a", "kind": "letter"}
        assert doc_full == {"_id": "a", "seq": 1, "kind": "letter"}

    @pytest.mark.describe(
        "test of custom_datatypes_in_reading APIOptions setting, async"
    )
    async def test_custom_datatypes_in_reading_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol_standard_dtypes = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=False),
            ),
        )
        acol_custom_dtypes = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=True),
            ),
        )
        the_dtime = datetime(2000, 1, 1, 10, 11, 12, 123000, tzinfo=timezone.utc)
        the_date = date(1998, 12, 31)

        # read path
        await async_empty_collection.insert_one(
            {
                "_id": "t0",
                "the_dtime": the_dtime,
                "the_date": the_date,
            },
        )
        doc_standard_dtypes = await acol_standard_dtypes.find_one({"_id": "t0"})
        doc_custom_dtypes = await acol_custom_dtypes.find_one({"_id": "t0"})

        assert doc_standard_dtypes is not None
        assert doc_custom_dtypes is not None
        dtime_standard_dtypes = doc_standard_dtypes["the_dtime"]
        dtime_custom_dtypes = doc_custom_dtypes["the_dtime"]
        date_standard_dtypes = doc_standard_dtypes["the_date"]
        date_custom_dtypes = doc_custom_dtypes["the_date"]

        assert isinstance(dtime_standard_dtypes, datetime)
        assert isinstance(dtime_custom_dtypes, DataAPITimestamp)
        assert (
            DataAPITimestamp.from_datetime(dtime_standard_dtypes) == dtime_custom_dtypes
        )

        assert dtime_custom_dtypes == DataAPITimestamp.from_datetime(
            dtime_standard_dtypes
        )
        assert isinstance(date_standard_dtypes, datetime)
        assert isinstance(date_custom_dtypes, DataAPITimestamp)
        assert (
            DataAPITimestamp.from_datetime(date_standard_dtypes) == date_custom_dtypes
        )
        assert date_custom_dtypes == DataAPITimestamp.from_datetime(
            date_standard_dtypes
        )

        # write path
        await async_empty_collection.delete_one({"_id": "t0"})
        await acol_standard_dtypes.insert_one(
            {"_id": "default_dtype_dt", "the_dtime": the_dtime}
        )
        # in write path, custom classes still accepted
        await acol_standard_dtypes.insert_one(
            {
                "_id": "default_dtype_ats",
                "the_dtime": DataAPITimestamp.from_datetime(the_dtime),
            }
        )
        await acol_custom_dtypes.insert_one(
            {"_id": "custom_dtype_dt", "the_dtime": the_dtime}
        )
        await acol_custom_dtypes.insert_one(
            {
                "_id": "custom_dtype_ats",
                "the_dtime": DataAPITimestamp.from_datetime(the_dtime),
            }
        )

        all_dates = [doc["the_dtime"] async for doc in async_empty_collection.find({})]
        assert len(all_dates) == 4
        assert len(set(all_dates)) == 1

    @pytest.mark.describe("test of find_one_and_replace, async")
    async def test_collection_find_one_and_replace_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection

        resp0000 = await acol.find_one_and_replace({"f": 0}, {"r": 1})
        assert resp0000 is None
        assert await acol.count_documents({}, upper_bound=100) == 0

        resp0001 = await acol.find_one_and_replace({"f": 0}, {"r": 1}, sort={"x": 1})
        assert resp0001 is None
        assert await acol.count_documents({}, upper_bound=100) == 0

        resp0010 = await acol.find_one_and_replace({"f": 0}, {"r": 1}, upsert=True)
        assert resp0010 is None
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        resp0011 = await acol.find_one_and_replace(
            {"f": 0}, {"r": 1}, upsert=True, sort={"x": 1}
        )
        assert resp0011 is None
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp0100 = await acol.find_one_and_replace({"f": 0}, {"r": 1})
        assert resp0100 is not None
        assert resp0100["f"] == 0
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp0101 = await acol.find_one_and_replace({"f": 0}, {"r": 1}, sort={"x": 1})
        assert resp0101 is not None
        assert resp0101["f"] == 0
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp0110 = await acol.find_one_and_replace({"f": 0}, {"r": 1}, upsert=True)
        assert resp0110 is not None
        assert resp0110["f"] == 0
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp0111 = await acol.find_one_and_replace(
            {"f": 0}, {"r": 1}, upsert=True, sort={"x": 1}
        )
        assert resp0111 is not None
        assert resp0111["f"] == 0
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        resp1000 = await acol.find_one_and_replace(
            {"f": 0}, {"r": 1}, return_document=ReturnDocument.AFTER
        )
        assert resp1000 is None
        assert await acol.count_documents({}, upper_bound=100) == 0

        resp1001 = await acol.find_one_and_replace(
            {"f": 0}, {"r": 1}, sort={"x": 1}, return_document=ReturnDocument.AFTER
        )
        assert resp1001 is None
        assert await acol.count_documents({}, upper_bound=100) == 0

        resp1010 = await acol.find_one_and_replace(
            {"f": 0}, {"r": 1}, upsert=True, return_document=ReturnDocument.AFTER
        )
        assert resp1010 is not None
        assert resp1010["r"] == 1
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        resp1011 = await acol.find_one_and_replace(
            {"f": 0},
            {"r": 1},
            upsert=True,
            sort={"x": 1},
            return_document=ReturnDocument.AFTER,
        )
        assert resp1011 is not None
        assert resp1011["r"] == 1
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp1100 = await acol.find_one_and_replace(
            {"f": 0}, {"r": 1}, return_document=ReturnDocument.AFTER
        )
        assert resp1100 is not None
        assert resp1100["r"] == 1
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp1101 = await acol.find_one_and_replace(
            {"f": 0}, {"r": 1}, sort={"x": 1}, return_document=ReturnDocument.AFTER
        )
        assert resp1101 is not None
        assert resp1101["r"] == 1
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp1110 = await acol.find_one_and_replace(
            {"f": 0}, {"r": 1}, upsert=True, return_document=ReturnDocument.AFTER
        )
        assert resp1110 is not None
        assert resp1110["r"] == 1
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp1111 = await acol.find_one_and_replace(
            {"f": 0},
            {"r": 1},
            upsert=True,
            sort={"x": 1},
            return_document=ReturnDocument.AFTER,
        )
        assert resp1111 is not None
        assert resp1111["r"] == 1
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        # projection
        await acol.insert_one({"f": 100, "name": "apple", "mode": "old"})
        resp_pr1 = await acol.find_one_and_replace(
            {"f": 100},
            {"f": 100, "name": "carrot", "mode": "replaced"},
            projection=["mode"],
            return_document=ReturnDocument.AFTER,
        )
        assert resp_pr1 is not None
        assert set(resp_pr1.keys()) == {"_id", "mode"}
        resp_pr2 = await acol.find_one_and_replace(
            {"f": 100},
            {"f": 100, "name": "turnip", "mode": "re-replaced"},
            projection={"name": False, "f": False, "_id": False},
            return_document=ReturnDocument.BEFORE,
        )
        assert resp_pr2 is not None
        assert set(resp_pr2.keys()) == {"mode"}
        await acol.delete_many({})

    @pytest.mark.describe("test of replace_one, async")
    async def test_collection_replace_one_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection

        result1 = await acol.replace_one(filter={"a": 1}, replacement={"b": 2})
        assert result1.update_info["n"] == 0
        assert result1.update_info["updatedExisting"] is False
        assert result1.update_info["nModified"] == 0
        assert "upserted" not in result1.update_info

        result2 = await acol.replace_one(
            filter={"a": 1}, replacement={"b": 2}, upsert=True
        )
        assert result2.update_info["n"] == 1
        assert result2.update_info["updatedExisting"] is False
        assert result2.update_info["nModified"] == 0
        assert "upserted" in result2.update_info

        result3 = await acol.replace_one(filter={"b": 2}, replacement={"c": 3})
        assert result3.update_info["n"] == 1
        assert result3.update_info["updatedExisting"] is True
        assert result3.update_info["nModified"] == 1
        assert "upserted" not in result3.update_info

        result4 = await acol.replace_one(
            filter={"c": 3}, replacement={"d": 4}, upsert=True
        )
        assert result4.update_info["n"] == 1
        assert result4.update_info["updatedExisting"] is True
        assert result4.update_info["nModified"] == 1
        assert "upserted" not in result4.update_info

        # test of sort
        await async_empty_collection.insert_many(
            [{"ts": 1, "seq": i} for i in [2, 0, 1]]
        )
        await async_empty_collection.replace_one(
            {"ts": 1}, {"ts": 1, "R": True}, sort={"seq": 1}
        )
        assert set(await async_empty_collection.distinct("seq", filter={"ts": 1})) == {
            1,
            2,
        }

    @pytest.mark.describe("test of replace_one with vectors, async")
    async def test_collection_replace_one_vector_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection
        await acol.insert_many(
            [
                {"tag": "h", "$vector": [10, 5]},
                {"tag": "v", "$vector": [2, 20]},
            ]
        )
        result = await acol.replace_one({}, {"new_doc": True}, sort={"$vector": [0, 1]})
        assert result.update_info["updatedExisting"]

        assert (await acol.find_one({"tag": "h"})) is not None

    @pytest.mark.describe("test of update_one, async")
    async def test_collection_update_one_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection

        result1 = await acol.update_one(filter={"a": 1}, update={"$set": {"b": 2}})
        assert result1.update_info["n"] == 0
        assert result1.update_info["updatedExisting"] is False
        assert result1.update_info["nModified"] == 0
        assert "upserted" not in result1.update_info

        result2 = await acol.update_one(
            filter={"a": 1}, update={"$set": {"b": 2}}, upsert=True
        )
        assert result2.update_info["n"] == 1
        assert result2.update_info["updatedExisting"] is False
        assert result2.update_info["nModified"] == 0
        assert "upserted" in result2.update_info

        result3 = await acol.update_one(filter={"b": 2}, update={"$set": {"c": 3}})
        assert result3.update_info["n"] == 1
        assert result3.update_info["updatedExisting"] is True
        assert result3.update_info["nModified"] == 1
        assert "upserted" not in result3.update_info

        result4 = await acol.update_one(
            filter={"c": 3}, update={"$set": {"d": 4}}, upsert=True
        )
        assert result4.update_info["n"] == 1
        assert result4.update_info["updatedExisting"] is True
        assert result4.update_info["nModified"] == 1
        assert "upserted" not in result4.update_info

        # test of sort
        await async_empty_collection.insert_many(
            [{"ts": 1, "seq": i} for i in [2, 0, 1]]
        )
        await async_empty_collection.update_one(
            {"ts": 1}, {"$set": {"U": True}}, sort={"seq": 1}
        )
        updated = await async_empty_collection.find_one({"U": True})
        assert updated is not None
        assert updated["seq"] == 0

    @pytest.mark.describe("test of update_many, async")
    async def test_collection_update_many_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection
        await acol.insert_many([{"a": 1, "seq": i} for i in range(4)])
        await acol.insert_many([{"a": 2, "seq": i} for i in range(2)])

        resp1 = await acol.update_many({"a": 1}, {"$set": {"n": 1}})
        assert resp1.update_info["n"] == 4
        assert resp1.update_info["updatedExisting"] is True
        assert resp1.update_info["nModified"] == 4
        assert "upserted" not in resp1.update_info

        resp2 = await acol.update_many({"a": 1}, {"$set": {"n": 2}}, upsert=True)
        assert resp2.update_info["n"] == 4
        assert resp2.update_info["updatedExisting"] is True
        assert resp2.update_info["nModified"] == 4
        assert "upserted" not in resp2.update_info

        resp3 = await acol.update_many({"a": 3}, {"$set": {"n": 3}})
        assert resp3.update_info["n"] == 0
        assert resp3.update_info["updatedExisting"] is False
        assert resp3.update_info["nModified"] == 0
        assert "upserted" not in resp3.update_info

        resp4 = await acol.update_many({"a": 3}, {"$set": {"n": 4}}, upsert=True)
        assert resp4.update_info["n"] == 1
        assert resp4.update_info["updatedExisting"] is False
        assert resp4.update_info["nModified"] == 0
        assert "upserted" in resp4.update_info

    @pytest.mark.describe("test of update_many, async")
    async def test_collection_paginated_update_many_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection
        await acol.insert_many([{"a": 1} for _ in range(50)])
        await acol.insert_many([{"a": 10} for _ in range(10)])

        um_result = await acol.update_many({"a": 1}, {"$set": {"b": 2}})
        assert um_result.update_info["n"] == 50
        assert um_result.update_info["updatedExisting"] is True
        assert um_result.update_info["nModified"] == 50
        assert "upserted" not in um_result.update_info
        assert "upsertedd" not in um_result.update_info
        assert await acol.count_documents({"b": 2}, upper_bound=100) == 50
        assert await acol.count_documents({}, upper_bound=100) == 60

    @pytest.mark.describe("test of collection find_one_and_delete, async")
    async def test_collection_find_one_and_delete_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        await async_empty_collection.insert_one({"doc": 1, "group": "A"})
        await async_empty_collection.insert_one({"doc": 2, "group": "B"})
        await async_empty_collection.insert_one({"doc": 3, "group": "A"})
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
            == 3
        )

        fo_result1 = await async_empty_collection.find_one_and_delete({"group": "A"})
        assert fo_result1 is not None
        assert set(fo_result1.keys()) == {"_id", "doc", "group"}
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
            == 2
        )

        fo_result2 = await async_empty_collection.find_one_and_delete(
            {"group": "B"}, projection=["doc"]
        )
        assert fo_result2 is not None
        assert set(fo_result2.keys()) == {"_id", "doc"}
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
            == 1
        )

        fo_result3 = await async_empty_collection.find_one_and_delete(
            {"group": "A"}, projection={"_id": False, "group": False}
        )
        assert fo_result3 is not None
        assert set(fo_result3.keys()) == {"doc"}
        assert (
            await async_empty_collection.count_documents(filter={}, upper_bound=100)
            == 0
        )

        fo_result4 = await async_empty_collection.find_one_and_delete({}, sort={"f": 1})
        assert fo_result4 is None

    @pytest.mark.describe("test of collection find_one_and_delete with vectors, async")
    async def test_collection_find_one_and_delete_vectors_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection
        await acol.insert_many(
            [
                {"tag": "h", "$vector": [10, 5]},
                {"tag": "v", "$vector": [2, 20]},
            ]
        )
        deleted = await acol.find_one_and_delete({}, sort={"$vector": [0, 1]})
        assert deleted is not None
        assert deleted["tag"] == "v"

    @pytest.mark.describe("test of find_one_and_update, async")
    async def test_collection_find_one_and_update_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        acol = async_empty_collection

        resp0000 = await acol.find_one_and_update({"f": 0}, {"$set": {"n": 1}})
        assert resp0000 is None
        assert await acol.count_documents({}, upper_bound=100) == 0

        resp0001 = await acol.find_one_and_update(
            {"f": 0}, {"$set": {"n": 1}}, sort={"x": 1}
        )
        assert resp0001 is None
        assert await acol.count_documents({}, upper_bound=100) == 0

        resp0010 = await acol.find_one_and_update(
            {"f": 0}, {"$set": {"n": 1}}, upsert=True
        )
        assert resp0010 is None
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        resp0011 = await acol.find_one_and_update(
            {"f": 0}, {"$set": {"n": 1}}, upsert=True, sort={"x": 1}
        )
        assert resp0011 is None
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp0100 = await acol.find_one_and_update({"f": 0}, {"$set": {"n": 1}})
        assert resp0100 is not None
        assert resp0100["f"] == 0
        assert "n" not in resp0100
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp0101 = await acol.find_one_and_update(
            {"f": 0}, {"$set": {"n": 1}}, sort={"x": 1}
        )
        assert resp0101 is not None
        assert resp0101["f"] == 0
        assert "n" not in resp0101
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp0110 = await acol.find_one_and_update(
            {"f": 0}, {"$set": {"n": 1}}, upsert=True
        )
        assert resp0110 is not None
        assert resp0110["f"] == 0
        assert "n" not in resp0110
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp0111 = await acol.find_one_and_update(
            {"f": 0}, {"$set": {"n": 1}}, upsert=True, sort={"x": 1}
        )
        assert resp0111 is not None
        assert resp0111["f"] == 0
        assert "n" not in resp0111
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        resp1000 = await acol.find_one_and_update(
            {"f": 0}, {"$set": {"n": 1}}, return_document=ReturnDocument.AFTER
        )
        assert resp1000 is None
        assert await acol.count_documents({}, upper_bound=100) == 0

        resp1001 = await acol.find_one_and_update(
            {"f": 0},
            {"$set": {"n": 1}},
            sort={"x": 1},
            return_document=ReturnDocument.AFTER,
        )
        assert resp1001 is None
        assert await acol.count_documents({}, upper_bound=100) == 0

        resp1010 = await acol.find_one_and_update(
            {"f": 0},
            {"$set": {"n": 1}},
            upsert=True,
            return_document=ReturnDocument.AFTER,
        )
        assert resp1010 is not None
        assert resp1010["n"] == 1
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        resp1011 = await acol.find_one_and_update(
            {"f": 0},
            {"$set": {"n": 1}},
            upsert=True,
            sort={"x": 1},
            return_document=ReturnDocument.AFTER,
        )
        assert resp1011 is not None
        assert resp1011["n"] == 1
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp1100 = await acol.find_one_and_update(
            {"f": 0}, {"$set": {"n": 1}}, return_document=ReturnDocument.AFTER
        )
        assert resp1100 is not None
        assert resp1100["n"] == 1
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp1101 = await acol.find_one_and_update(
            {"f": 0},
            {"$set": {"n": 1}},
            sort={"x": 1},
            return_document=ReturnDocument.AFTER,
        )
        assert resp1101 is not None
        assert resp1101["n"] == 1
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp1110 = await acol.find_one_and_update(
            {"f": 0},
            {"$set": {"n": 1}},
            upsert=True,
            return_document=ReturnDocument.AFTER,
        )
        assert resp1110 is not None
        assert resp1110["n"] == 1
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        await acol.insert_one({"f": 0})
        resp1111 = await acol.find_one_and_update(
            {"f": 0},
            {"$set": {"n": 1}},
            upsert=True,
            sort={"x": 1},
            return_document=ReturnDocument.AFTER,
        )
        assert resp1111 is not None
        assert resp1111["n"] == 1
        assert await acol.count_documents({}, upper_bound=100) == 1
        await acol.delete_many({})

        # projection
        await acol.insert_one({"f": 100, "name": "apple", "mode": "old"})
        resp_pr1 = await acol.find_one_and_update(
            {"f": 100},
            {"$unset": {"mode": ""}},
            projection=["mode", "f"],
            return_document=ReturnDocument.AFTER,
        )
        assert resp_pr1 is not None
        assert set(resp_pr1.keys()) == {"_id", "f"}
        resp_pr2 = await acol.find_one_and_update(
            {"f": 100},
            {"$set": {"mode": "re-replaced"}},
            projection={"name": False, "_id": False},
            return_document=ReturnDocument.BEFORE,
        )
        assert resp_pr2 is not None
        assert set(resp_pr2.keys()) == {"f"}
        await acol.delete_many({})

    @pytest.mark.describe("test of the various ids in the document id field, async")
    async def test_collection_ids_as_doc_id_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        types_and_ids = {
            "uuid1": UUID("8ccd6ff8-e61b-11ee-a2fc-7df4a8c4164b"),
            "uuid3": UUID("6fa459ea-ee8a-3ca4-894e-db77e160355e"),
            "uuid4": UUID("4f16cba8-1115-43ab-aa39-3a9c29f37db5"),
            "uuid5": UUID("886313e1-3b8a-5372-9b90-0c9aee199e5d"),
            "uuid6": UUID("1eee61b9-8f2d-69ad-8ebb-5054d2a1a2c0"),
            "uuid7": UUID("018e57e5-f586-7ed6-be55-6b0de3041116"),
            "uuid8": UUID("018e57e5-fbcd-8bd4-b794-be914f2c4c85"),
            "objectid": ObjectId("65f9cfa0d7fabb3f255c25a1"),
        }

        await async_empty_collection.insert_many(
            [
                {"_id": t_id, "id_type": t_id_type}
                for t_id_type, t_id in types_and_ids.items()
            ]
        )

        for t_id_type, t_id in types_and_ids.items():
            this_doc = await async_empty_collection.find_one(
                {"_id": t_id},
                projection={"id_type": True},
            )
            assert this_doc is not None
            assert this_doc["id_type"] == t_id_type

    @pytest.mark.describe(
        "test of ids in various parameters of various DML methods, async"
    )
    async def test_collection_ids_throughout_dml_methods_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        types_and_ids = {
            "uuid1": UUID("8ccd6ff8-e61b-11ee-a2fc-7df4a8c4164b"),
            "uuid3": UUID("6fa459ea-ee8a-3ca4-894e-db77e160355e"),
            "uuid4": UUID("4f16cba8-1115-43ab-aa39-3a9c29f37db5"),
            "uuid5": UUID("886313e1-3b8a-5372-9b90-0c9aee199e5d"),
            "uuid6": UUID("1eee61b9-8f2d-69ad-8ebb-5054d2a1a2c0"),
            "uuid7": UUID("018e57e5-f586-7ed6-be55-6b0de3041116"),
            "uuid8": UUID("018e57e5-fbcd-8bd4-b794-be914f2c4c85"),
            "objectid": ObjectId("65f9cfa0d7fabb3f255c25a1"),
        }
        wide_document = {
            "all_ids": types_and_ids,
            "_id": 0,
            "name": "wide_document",
            "touched_times": 0,
        }
        await async_empty_collection.insert_one(wide_document)

        full_doc = await async_empty_collection.find_one({})
        assert full_doc == wide_document

        for t_id_type, t_id in types_and_ids.items():
            doc = await async_empty_collection.find_one(
                {f"all_ids.{t_id_type}": t_id}, projection={"name": True}
            )
            assert doc is not None
            assert doc["name"] == "wide_document"

        for upd_index, (t_id_type, t_id) in enumerate(types_and_ids.items()):
            updated_doc = await async_empty_collection.find_one_and_update(
                {f"all_ids.{t_id_type}": t_id},
                {"$inc": {"touched_times": 1}},
                return_document=ReturnDocument.AFTER,
            )
            assert updated_doc is not None
            assert updated_doc["touched_times"] == upd_index + 1

        await async_empty_collection.delete_one({"_id": 0})

        await async_empty_collection.insert_many(
            [{"_id": t_id} for t_id in types_and_ids.values()]
        )

        count = await async_empty_collection.count_documents({}, upper_bound=20)
        assert count == len(types_and_ids)

        for del_index, t_id in enumerate(types_and_ids.values()):
            del_result = await async_empty_collection.delete_one({"_id": t_id})
            assert del_result.deleted_count == 1
            count = await async_empty_collection.count_documents({}, upper_bound=20)
            assert count == len(types_and_ids) - del_index - 1

    @pytest.mark.describe("test of inserting various data types in a collection, async")
    async def test_collection_datatype_insertability_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        d_a_ts = DataAPITimestamp.from_string("1999-11-30T00:00:00.000Z")
        tz = timezone.utc
        at_document = {
            "_id": UUID("1f009012-ff61-646d-8c70-5d87cfbdee0b"),
            "int": 1,
            "string": "string",
            "float": 1.234,
            "dataapidate": DataAPIDate.from_string("1999-11-30"),
            "date": date(1999, 11, 30),
            "dataapitimestamp": d_a_ts,
            "datetime": d_a_ts.to_datetime(tz=tz),
            "dataapimap": DataAPIMap([("k", "v")]),
            "map": {"k": "v"},
        }
        at_expected = {
            **at_document,
            **{
                "dataapidate": d_a_ts,
                "date": d_a_ts,
                "datetime": d_a_ts,
                "dataapimap": {"k": "v"},
            },
        }

        ior = await async_empty_collection.insert_one(at_document)
        assert ior.inserted_id == at_document["_id"]

        read_document = await async_empty_collection.find_one({})
        assert read_document is not None
        date_checkfields = {"date", "dataapidate"}
        ok_read = {k: v for k, v in read_document.items() if k not in date_checkfields}
        ok_expected = {
            k: v for k, v in at_expected.items() if k not in date_checkfields
        }
        assert ok_read == ok_expected
        # irreducible approximate check on 'just-date' fields
        assert isinstance(read_document["date"], DataAPITimestamp)
        assert isinstance(read_document["dataapidate"], DataAPITimestamp)
        one_day_ms = 1000 * 86400
        assert (
            abs(read_document["date"].timestamp_ms - at_expected["date"].timestamp_ms)  # type: ignore[attr-defined]
            < one_day_ms
        )
        assert (
            abs(
                read_document["dataapidate"].timestamp_ms
                - at_expected["dataapidate"].timestamp_ms  # type: ignore[attr-defined]
            )
            < one_day_ms
        )

    @pytest.mark.describe(
        "test of collection binary-encoding vectors everywhere, async"
    )
    async def test_collection_binencvectors_everywhere_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        binenc_options = APIOptions(
            serdes_options=SerdesOptions(binary_encode_vectors=True),
        )
        binenc_acoll = async_empty_collection.with_options(api_options=binenc_options)

        # Using DataAPIVector (the serializer binary-encodes all of these)
        await binenc_acoll.insert_one({"_id": "0", "$vector": DataAPIVector([1, 1])})
        await binenc_acoll.insert_many(
            [{"_id": "X1", "$vector": DataAPIVector([1, 1])}]
        )
        await binenc_acoll.update_one(
            {"_id": "0"}, {"$set": {"$vector": DataAPIVector([0, 1])}}, upsert=True
        )
        await binenc_acoll.update_one(
            {"_id": "X2"},
            {"$setOnInsert": {"$vector": DataAPIVector([0, 1])}},
            upsert=True,
        )
        await binenc_acoll.update_one(
            {},
            sort={"$vector": DataAPIVector([-1, 0])},
            update={"$set": {"oc": "ooo0"}},
            upsert=True,
        )
        await binenc_acoll.update_many(
            {"_id": "0"}, {"$set": {"$vector": DataAPIVector([0, 1])}}, upsert=True
        )
        await binenc_acoll.update_many(
            {"_id": "X3"},
            {"$setOnInsert": {"$vector": DataAPIVector([0, 1])}},
            upsert=True,
        )
        await binenc_acoll.replace_one(
            {"_id": "0"}, {"$vector": DataAPIVector([1, 2])}, upsert=True
        )
        await binenc_acoll.replace_one(
            {}, {"oc": "ooo1"}, sort={"$vector": DataAPIVector([4, 1])}, upsert=True
        )
        await binenc_acoll.delete_one({}, sort={"$vector": DataAPIVector([5, 4])})
        await binenc_acoll.find_one(sort={"$vector": DataAPIVector([5, 4])})
        await binenc_acoll.find(sort={"$vector": DataAPIVector([5, 4])}).to_list()
        await binenc_acoll.find_and_rerank(
            {},
            sort={"$hybrid": {"$vector": DataAPIVector([-1, -2]), "$lexical": "bla"}},
            rerank_query="bla",
            rerank_on="$lexical",
        ).to_list()
        await binenc_acoll.find_one_and_replace(
            {}, {"$vector": DataAPIVector([5, 4])}, upsert=True
        )
        await binenc_acoll.find_one_and_replace(
            {}, {"oc": "ooo2"}, sort={"$vector": DataAPIVector([9, 1])}, upsert=True
        )
        await binenc_acoll.find_one_and_delete(
            {}, sort={"$vector": DataAPIVector([9, 1])}
        )
        await binenc_acoll.find_one_and_update(
            {"_id": "0"}, {"$set": {"$vector": DataAPIVector([5, 2])}}, upsert=True
        )
        await binenc_acoll.find_one_and_update(
            {"_id": "X4"},
            {"$setOnInsert": {"$vector": DataAPIVector([5, 2])}},
            upsert=True,
        )
        await binenc_acoll.find_one_and_update(
            {},
            {"$set": {"oc": "ooo3"}},
            sort={"$vector": DataAPIVector([7, 1])},
            upsert=True,
        )

        # Using a plain list (the serializer binary-encodes all of these anyway)
        await binenc_acoll.insert_one({"_id": "L0", "$vector": [1, 1]})
        await binenc_acoll.insert_many([{"_id": "LX1", "$vector": [1, 1]}])
        await binenc_acoll.update_one(
            {"_id": "L0"}, {"$set": {"$vector": [0, 1]}}, upsert=True
        )
        await binenc_acoll.update_one(
            {"_id": "LX2"}, {"$setOnInsert": {"$vector": [0, 1]}}, upsert=True
        )
        await binenc_acoll.update_one(
            {}, sort={"$vector": [-1, 0]}, update={"$set": {"oc": "qqq0"}}, upsert=True
        )
        await binenc_acoll.update_many(
            {"_id": "L0"}, {"$set": {"$vector": [0, 1]}}, upsert=True
        )
        await binenc_acoll.update_many(
            {"_id": "LX3"}, {"$setOnInsert": {"$vector": [0, 1]}}, upsert=True
        )
        await binenc_acoll.replace_one({"_id": "L0"}, {"$vector": [1, 2]}, upsert=True)
        await binenc_acoll.replace_one(
            {}, {"oc": "qqq1"}, sort={"$vector": [4, 1]}, upsert=True
        )
        await binenc_acoll.delete_one({}, sort={"$vector": [5, 4]})
        await binenc_acoll.find_one(sort={"$vector": [5, 4]})
        await binenc_acoll.find(sort={"$vector": [5, 4]}).to_list()
        await binenc_acoll.find_and_rerank(
            {},
            sort={"$hybrid": {"$vector": [-1, -2], "$lexical": "bla"}},
            rerank_query="bla",
            rerank_on="$lexical",
        ).to_list()
        await binenc_acoll.find_one_and_replace({}, {"$vector": [5, 4]}, upsert=True)
        await binenc_acoll.find_one_and_replace(
            {}, {"oc": "qqq2"}, sort={"$vector": [9, 1]}, upsert=True
        )
        await binenc_acoll.find_one_and_delete({}, sort={"$vector": [9, 1]})
        await binenc_acoll.find_one_and_update(
            {"_id": "L0"}, {"$set": {"$vector": [5, 2]}}, upsert=True
        )
        await binenc_acoll.find_one_and_update(
            {"_id": "LX4"}, {"$setOnInsert": {"$vector": [5, 2]}}, upsert=True
        )
        await binenc_acoll.find_one_and_update(
            {}, {"$set": {"oc": "qqq3"}}, sort={"$vector": [7, 1]}, upsert=True
        )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pickle

import pytest

from astrapy.data_types import DataAPIMap


class TestDataAPIMap:
    @pytest.mark.describe("test of table map usage with hashables")
    def test_dataapimap_hashables(self) -> None:
        mp0: DataAPIMap[int, int] = DataAPIMap()
        assert mp0 == DataAPIMap()
        assert dict(mp0) == {}
        assert mp0 == {}
        # identity/equality
        items1 = [(1, "a"), (2, "b"), (3, "c")]
        mp1 = DataAPIMap(items1)
        assert mp1 == DataAPIMap(items1)
        assert mp1 == DataAPIMap(items1 + [(2, "b")])
        assert mp1 == DataAPIMap(items1[2:] + items1[:2])
        assert mp1 != DataAPIMap(items1[1:] + [(1, "z")])
        assert mp1 != DataAPIMap(items1[1:] + [(9, "z")])
        assert dict(items1) == mp1
        assert mp1 == dict(items1)
        assert mp1 == dict(items1 + [(2, "b")])
        assert mp1 == dict(items1[2:] + items1[:2])
        assert mp1 != dict(items1[1:] + [(1, "z")])
        assert mp1 != dict(items1[1:] + [(9, "z")])
        # map operations
        assert list(mp1.keys()) == [1, 2, 3]
        assert list(mp1.values()) == ["a", "b", "c"]
        assert mp1[2] == "b"
        with pytest.raises(KeyError):
            mp1[4]
        assert list(mp1.items()) == items1
        assert len(mp1) == len(items1)
        assert list(mp1) == [1, 2, 3]
        # bool
        assert not bool(mp0)
        assert bool(mp1)

    @pytest.mark.describe("test of table map usage with non-hashables")
    def test_dataapimap_nonhashables(self) -> None:
        mp0: DataAPIMap[list[int], int] = DataAPIMap()
        assert mp0 == DataAPIMap()
        assert dict(mp0) == {}
        # identity/equality
        items1 = [([1], "a"), ([2], "b"), ([3], "c")]
        mp1 = DataAPIMap(items1)
        assert mp1 == DataAPIMap(items1)
        assert mp1 == DataAPIMap(items1 + [([2], "b")])
        assert mp1 == DataAPIMap(items1[2:] + items1[:2])
        assert mp1 != DataAPIMap(items1[1:] + [([1], "z")])
        assert mp1 != DataAPIMap(items1[1:] + [([9], "z")])
        # map operations
        assert list(mp1.keys()) == [[1], [2], [3]]
        assert list(mp1.values()) == ["a", "b", "c"]
        assert mp1[[2]] == "b"
        with pytest.raises(KeyError):
            mp1[[4]]
        assert list(mp1.items()) == items1
        assert len(mp1) == len(items1)
        assert list(mp1) == [[1], [2], [3]]

    @pytest.mark.describe("test pickling of DataAPIMap")
    def test_dataapimap_pickle(self) -> None:
        the_map = DataAPIMap([("key1", 1), (None, 2), ("key3", {"a": 1})])
        assert pickle.loads(pickle.dumps(the_map)) == the_map



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import datetime
import pickle

import pytest

from astrapy.data_types import DataAPITime


class TestDataAPITime:
    @pytest.mark.describe("test of time type, errors in parsing from string")
    def test_dataapitime_parse_errors(self) -> None:
        # empty, faulty, misformatted
        with pytest.raises(ValueError):
            DataAPITime.from_string("")
        with pytest.raises(ValueError):
            DataAPITime.from_string("boom")

        with pytest.raises(ValueError):
            DataAPITime.from_string("12:34:56:21")
        with pytest.raises(ValueError):
            DataAPITime.from_string("+12:34:56")
        with pytest.raises(ValueError):
            DataAPITime.from_string("12:+34:56")
        with pytest.raises(ValueError):
            DataAPITime.from_string("12:34:56.")
        with pytest.raises(ValueError):
            DataAPITime.from_string("12:34:56X")
        with pytest.raises(ValueError):
            DataAPITime.from_string("X12:34:56")
        with pytest.raises(ValueError):
            DataAPITime.from_string("12:34X:56")

        with pytest.raises(ValueError):
            DataAPITime.from_string("24:00:00")
        with pytest.raises(ValueError):
            DataAPITime.from_string("00:60:00")
        with pytest.raises(ValueError):
            DataAPITime.from_string("00:00:60")

        DataAPITime.from_string("12:23")
        DataAPITime.from_string("00:00:00")
        DataAPITime.from_string("23:00:00")
        DataAPITime.from_string("00:59:00")
        DataAPITime.from_string("00:00:00.123456789")
        DataAPITime.from_string("00:00:00.123")
        DataAPITime.from_string("01:02:03.123456789")
        DataAPITime.from_string("01:02:03.123")

    @pytest.mark.describe("test of time type, lifecycle")
    def test_dataapitime_lifecycle(self) -> None:
        tint = DataAPITime.from_string("02:03:04")
        tint_exp = DataAPITime(hour=2, minute=3, second=4)
        assert tint == tint_exp
        assert tint == DataAPITime.from_string("2:3:4")
        assert tint == DataAPITime.from_string("0002:003:000004")
        py_tint = datetime.time(2, 3, 4)
        assert DataAPITime.from_time(py_tint) == tint
        assert tint.to_time() == py_tint
        repr(tint)
        str(tint)
        tint.to_string()

        tfra = DataAPITime.from_string("02:03:04.9876")
        tfra_exp = DataAPITime(hour=2, minute=3, second=4, nanosecond=987600000)
        assert tfra == tfra_exp
        assert tfra == DataAPITime.from_string("2:3:4.9876")
        assert tfra == DataAPITime.from_string("0002:003:000004.987600000")
        py_tfra = datetime.time(2, 3, 4, 987600)
        assert DataAPITime.from_time(py_tfra) == tfra
        assert tfra.to_time() == py_tfra
        repr(tfra)
        str(tfra)
        tfra.to_string()

        tfra1 = DataAPITime(1, 2, 3, 123)
        tfra1p = DataAPITime(1, 2, 3, 12)
        tfra2p = DataAPITime(1, 2, 3, 12345)
        tfra2 = DataAPITime(1, 2, 3, 123456)
        tfra3p = DataAPITime(1, 2, 3, 12345678)
        tfra3 = DataAPITime(1, 2, 3, 123456789)
        assert DataAPITime.from_string(tfra1.to_string()) == tfra1
        assert DataAPITime.from_string(tfra1p.to_string()) == tfra1p
        assert DataAPITime.from_string(tfra2p.to_string()) == tfra2p
        assert DataAPITime.from_string(tfra2.to_string()) == tfra2
        assert DataAPITime.from_string(tfra3p.to_string()) == tfra3p
        assert DataAPITime.from_string(tfra3.to_string()) == tfra3

        t1 = DataAPITime(1, 2, 3, 30000)
        t2 = DataAPITime(1, 2, 3, 45000)
        py_t1 = t1.to_time()
        py_t2 = t2.to_time()
        assert t1 < t2
        assert t1 <= t2
        assert t2 > t1
        assert t2 >= t1
        assert py_t1 < t2
        assert py_t1 <= t2
        assert py_t2 > t1
        assert py_t2 >= t1
        assert t1 < py_t2
        assert t1 <= py_t2
        assert t2 > py_t1
        assert t2 >= py_t1

    @pytest.mark.describe("test pickling of DataAPITime")
    def test_dataapitime_pickle(self) -> None:
        the_time = DataAPITime(12, 34, 56, 789)
        assert pickle.loads(pickle.dumps(the_time)) == the_time



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import sys
from typing import Any

import pytest

from astrapy import Database
from astrapy.authentication import AWSEmbeddingHeadersProvider, EmbeddingHeadersProvider
from astrapy.exceptions import CollectionInsertManyException, DataAPIResponseException
from astrapy.info import CollectionDefinition, VectorServiceOptions

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from ..conftest import IS_ASTRA_DB
from ..vectorize_models import live_test_models


def enabled_vectorize_models(auth_type: str) -> list[Any]:
    """
    This actually returns a list of `_pytest.mark.structures.ParameterSet` instances,
    each wrapping a dict with the needed info to test the model

    The tested models can also be further restricted by:
        EMBEDDING_MODEL_TAGS="tag1,tag2..." [...pytest command...]
    where `tag` = "provider/model/auth_type/[0 or f]"
    """
    all_test_models = list(live_test_models())
    all_model_ids: list[str] = [
        str(model_desc["model_tag"]) for model_desc in all_test_models
    ]
    #
    at_test_models = [
        test_model
        for test_model in all_test_models
        if test_model["auth_type_name"] == auth_type
    ]
    at_model_ids: list[str] = [
        str(model_desc["model_tag"]) for model_desc in at_test_models
    ]
    at_chosen_models: list[Any] = []
    if "EMBEDDING_MODEL_TAGS" in os.environ:
        whitelisted_models = [
            _wmd.strip()
            for _wmd in os.environ["EMBEDDING_MODEL_TAGS"].split(",")
            if _wmd.strip()
        ]
    else:
        whitelisted_models = at_model_ids
    unknown_wl_models = [wmd for wmd in whitelisted_models if wmd not in all_model_ids]
    if unknown_wl_models:
        raise ValueError(f"Unknown whitelisted model(s): {','.join(unknown_wl_models)}")

    for model in at_test_models:
        markers = []
        # provider exclusion logic applied here:
        if model["model_tag"] not in whitelisted_models:
            markers.append(pytest.mark.skip(reason="model not whitelisted"))
        at_chosen_models.append(pytest.param(model, marks=markers))
    return at_chosen_models


class TestVectorizeProviders:
    @pytest.mark.parametrize(
        "testable_vectorize_model",
        enabled_vectorize_models(auth_type="HEADER"),
        ids=[
            str(model["model_tag"])
            for model in live_test_models()
            if model["auth_type_name"] == "HEADER"
        ],
    )
    @pytest.mark.describe(
        "test of vectorize collection basic usage with header key, sync"
    )
    def test_vectorize_usage_auth_type_header_sync(
        self,
        sync_database: Database,
        testable_vectorize_model: dict[str, Any],
    ) -> None:
        simple_tag = testable_vectorize_model["simple_tag"].lower()
        # switch betewen header providers according to what is needed
        # For the time being this is necessary on HEADER only
        embedding_api_key: str | EmbeddingHeadersProvider
        at_tokens = testable_vectorize_model["auth_type_tokens"]
        at_token_lnames = {tk.accepted.lower() for tk in at_tokens}
        if at_token_lnames == {"x-embedding-api-key"}:
            embedding_api_key = os.environ[
                f"HEADER_EMBEDDING_API_KEY_{testable_vectorize_model['secret_tag']}"
            ]
        elif at_token_lnames == {"x-embedding-access-id", "x-embedding-secret-id"}:
            embedding_api_key = AWSEmbeddingHeadersProvider(
                embedding_access_id=os.environ[
                    f"HEADER_EMBEDDING_ACCESS_ID_{testable_vectorize_model['secret_tag']}"
                ],
                embedding_secret_id=os.environ[
                    f"HEADER_EMBEDDING_SECRET_ID_{testable_vectorize_model['secret_tag']}"
                ],
            )
        else:
            raise ValueError(
                f"Unsupported auth type tokens for {testable_vectorize_model['model_tag']}"
            )
        dimension = testable_vectorize_model.get("dimension")
        service_options = testable_vectorize_model["service_options"]

        db = sync_database
        collection_name = f"vec_h_{simple_tag}"
        # create with header
        try:
            collection = db.create_collection(
                collection_name,
                definition=(
                    CollectionDefinition.builder()
                    .set_vector_dimension(dimension)
                    .set_vector_metric("cosine")
                    .set_vector_service(service_options)
                    .build()
                ),
                embedding_api_key=embedding_api_key,
            )
            # this is to cope with the Data API normalizing options
            expected_service_options = VectorServiceOptions(
                provider=service_options.provider,
                model_name=testable_vectorize_model.get(
                    "expected_model_name", service_options.model_name
                ),
                parameters=service_options.parameters,
            )
            # test service back from collection info
            c_descriptors = [
                cd for cd in db.list_collections() if cd.name == collection_name
            ]
            assert len(c_descriptors) == 1
            c_descriptor = c_descriptors[0]
            assert c_descriptor.definition is not None
            assert c_descriptor.definition.vector is not None
            assert c_descriptor.definition.vector.service == expected_service_options
            # put entries
            test_assets = testable_vectorize_model["test_assets"]
            if testable_vectorize_model["use_insert_one"]:
                for test_sample_tag, test_sample_text in test_assets["samples"]:
                    collection.insert_one(
                        {"tag": test_sample_tag, "$vectorize": test_sample_text},
                    )
                # also test for an error if inserting many
                with pytest.raises(CollectionInsertManyException):
                    collection.insert_many(
                        [
                            {"tag": tag, "$vectorize": text}
                            for tag, text in test_assets["samples"]
                        ],
                    )
            else:
                collection.insert_many(
                    [
                        {"tag": tag, "$vectorize": text}
                        for tag, text in test_assets["samples"]
                    ],
                )
            # instantiate with header
            collection_i = db.get_collection(
                collection_name,
                embedding_api_key=embedding_api_key,
            )
            # run ANN and check results
            hits = [
                document["tag"]
                for document in collection_i.find(
                    sort={"$vectorize": test_assets["probe"]["text"]},
                    limit=len(test_assets["probe"]["expected"]),
                    projection=["tag"],
                )
            ]
            assert hits == test_assets["probe"]["expected"]
        finally:
            db.drop_collection(collection_name)

    @pytest.mark.skipif(not IS_ASTRA_DB, reason="Not supported outside of Astra DB")
    @pytest.mark.parametrize(
        "testable_vectorize_model",
        enabled_vectorize_models(auth_type="NONE"),
        ids=[
            str(model["model_tag"])
            for model in live_test_models()
            if model["auth_type_name"] == "NONE"
        ],
    )
    @pytest.mark.describe("test of vectorize collection basic usage with no auth, sync")
    def test_vectorize_usage_auth_type_none_sync(
        self,
        sync_database: Database,
        testable_vectorize_model: dict[str, Any],
    ) -> None:
        simple_tag = testable_vectorize_model["simple_tag"].lower()
        dimension = testable_vectorize_model.get("dimension")
        service_options = testable_vectorize_model["service_options"]

        db = sync_database
        collection_name = f"vec_n_{simple_tag}"
        # create with header
        try:
            collection = db.create_collection(
                collection_name,
                definition=(
                    CollectionDefinition.builder()
                    .set_vector_dimension(dimension)
                    .set_vector_metric("cosine")
                    .set_vector_service(service_options)
                    .build()
                ),
            )
            # this is to cope with the Data API normalizing options
            expected_service_options = VectorServiceOptions(
                provider=service_options.provider,
                model_name=testable_vectorize_model.get(
                    "expected_model_name", service_options.model_name
                ),
                parameters=service_options.parameters,
            )
            # test service back from collection info
            c_descriptors = [
                cd for cd in db.list_collections() if cd.name == collection_name
            ]
            assert len(c_descriptors) == 1
            c_descriptor = c_descriptors[0]
            assert c_descriptor.definition is not None
            assert c_descriptor.definition.vector is not None
            assert c_descriptor.definition.vector.service == expected_service_options
            # put entries
            test_assets = testable_vectorize_model["test_assets"]
            if testable_vectorize_model["use_insert_one"]:
                for test_sample_tag, test_sample_text in test_assets["samples"]:
                    collection.insert_one(
                        {"tag": test_sample_tag, "$vectorize": test_sample_text},
                    )
            else:
                collection.insert_many(
                    [
                        {"tag": tag, "$vectorize": text}
                        for tag, text in test_assets["samples"]
                    ],
                )
            # instantiate collection
            collection_i = db.get_collection(
                collection_name,
            )
            # run ANN and check results
            hits = [
                document["tag"]
                for document in collection_i.find(
                    sort={"$vectorize": test_assets["probe"]["text"]},
                    limit=len(test_assets["probe"]["expected"]),
                    projection=["tag"],
                )
            ]
            assert hits == test_assets["probe"]["expected"]
        finally:
            db.drop_collection(collection_name)

    @pytest.mark.skipif(not IS_ASTRA_DB, reason="Not supported outside of Astra DB")
    @pytest.mark.parametrize(
        "testable_vectorize_model",
        enabled_vectorize_models(auth_type="SHARED_SECRET"),
        ids=[
            str(model["model_tag"])
            for model in live_test_models()
            if model["auth_type_name"] == "SHARED_SECRET"
        ],
    )
    @pytest.mark.describe(
        "test of vectorize collection basic usage with shared_secret, sync"
    )
    def test_vectorize_usage_auth_type_shared_secret_sync(
        self,
        sync_database: Database,
        testable_vectorize_model: dict[str, Any],
    ) -> None:
        simple_tag = testable_vectorize_model["simple_tag"].lower()
        secret_tag = testable_vectorize_model["secret_tag"]
        dimension = testable_vectorize_model.get("dimension")
        base_service_options = testable_vectorize_model["service_options"]

        db = sync_database
        collection_name = f"vec_s_{simple_tag}"
        # to create the collection, prepare a service_options with the auth info
        service_options = VectorServiceOptions(
            provider=base_service_options.provider,
            model_name=base_service_options.model_name,
            authentication={
                "providerKey": f"SHARED_SECRET_EMBEDDING_API_KEY_{secret_tag}",
            },
            parameters=base_service_options.parameters,
        )
        # this is to cope with the Data API normalizing options
        expected_service_options = VectorServiceOptions(
            provider=base_service_options.provider,
            model_name=testable_vectorize_model.get(
                "expected_model_name", service_options.model_name
            ),
            authentication={
                "providerKey": f"SHARED_SECRET_EMBEDDING_API_KEY_{secret_tag}.providerKey",
            },
            parameters=base_service_options.parameters,
        )
        try:
            collection = db.create_collection(
                collection_name,
                definition=(
                    CollectionDefinition.builder()
                    .set_vector_dimension(dimension)
                    .set_vector_metric("cosine")
                    .set_vector_service(service_options)
                    .build()
                ),
            )
            # test service back from collection info
            c_descriptors = [
                cd for cd in db.list_collections() if cd.name == collection_name
            ]
            assert len(c_descriptors) == 1
            c_descriptor = c_descriptors[0]
            assert c_descriptor.definition is not None
            assert c_descriptor.definition.vector is not None
            assert c_descriptor.definition.vector.service == expected_service_options
            # put entries
            test_assets = testable_vectorize_model["test_assets"]
            if testable_vectorize_model["use_insert_one"]:
                for test_sample_tag, test_sample_text in test_assets["samples"]:
                    collection.insert_one(
                        {"tag": test_sample_tag, "$vectorize": test_sample_text},
                    )
            else:
                collection.insert_many(
                    [
                        {"tag": tag, "$vectorize": text}
                        for tag, text in test_assets["samples"]
                    ],
                )
            # instantiate collection
            collection_i = db.get_collection(
                collection_name,
            )
            # run ANN and check results
            hits = [
                document["tag"]
                for document in collection_i.find(
                    sort={"$vectorize": test_assets["probe"]["text"]},
                    limit=len(test_assets["probe"]["expected"]),
                    projection=["tag"],
                )
            ]
            assert hits == test_assets["probe"]["expected"]

            # test that header overrides shared secret:
            faulty_collection_i = db.get_collection(
                collection_name,
                embedding_api_key="clearly-not-a-working-secret",
            )
            with pytest.raises(DataAPIResponseException):
                faulty_collection_i.find_one(sort={"$vectorize": "Breaking sentence"})
        finally:
            db.drop_collection(collection_name)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.data.utils.collection_converters import (
    postprocess_collection_response,
    preprocess_collection_payload,
)
from astrapy.utils.api_commander import APICommander

from ..collection_decimal_support_assets import (
    BASELINE_OBJ,
    S_OPTS_NO_DECS,
    S_OPTS_OK_DECS,
    WDECS_OBJ,
    is_decimal_super,
)


class TestCollectionDecimalSupportUnit:
    @pytest.mark.describe(
        "test of decimals not supported by default in collection codec paths"
    )
    def test_decimalsupport_collection_codecpath_defaultsettings(self) -> None:
        # write path with baseline
        baseline_fully_encoded = APICommander._decimal_unaware_encode_payload(
            preprocess_collection_payload(
                BASELINE_OBJ,
                options=S_OPTS_NO_DECS,
            )
        )
        # read path back to it
        assert baseline_fully_encoded is not None
        baseline_obj_2 = postprocess_collection_response(
            APICommander._decimal_unaware_parse_json_response(
                baseline_fully_encoded,
            ),
            options=S_OPTS_NO_DECS,
        )
        # this must match exactly (baseline)
        assert BASELINE_OBJ == baseline_obj_2

        # write path with decimals should error instead
        with pytest.raises(TypeError):
            APICommander._decimal_unaware_encode_payload(
                preprocess_collection_payload(
                    WDECS_OBJ,
                    options=S_OPTS_NO_DECS,
                )
            )

    @pytest.mark.describe(
        "test of decimals supported in collection codec paths if set to do so"
    )
    def test_decimalsupport_collection_codecpath_decimalsettings(self) -> None:
        # write path with baseline
        baseline_fully_encoded = APICommander._decimal_aware_encode_payload(
            preprocess_collection_payload(
                BASELINE_OBJ,
                options=S_OPTS_OK_DECS,
            )
        )
        # read path back to it
        assert baseline_fully_encoded is not None
        baseline_obj_2 = postprocess_collection_response(
            APICommander._decimal_aware_parse_json_response(
                baseline_fully_encoded,
            ),
            options=S_OPTS_OK_DECS,
        )
        # the re-read object must "be more-or-equally Decimal" than the source
        # but otherwise coincide
        assert is_decimal_super(baseline_obj_2, BASELINE_OBJ)

        # write path with decimals
        wdecs_fully_encoded = APICommander._decimal_aware_encode_payload(
            preprocess_collection_payload(
                WDECS_OBJ,
                options=S_OPTS_OK_DECS,
            )
        )
        # read path back to it
        assert wdecs_fully_encoded is not None
        wdecs_2 = postprocess_collection_response(
            APICommander._decimal_aware_parse_json_response(
                wdecs_fully_encoded,
            ),
            options=S_OPTS_OK_DECS,
        )
        # the re-read object must "be more-or-equally Decimal" than the source
        # but otherwise coincide
        assert is_decimal_super(wdecs_2, WDECS_OBJ)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from preprocess_env import (
    ASTRA_DB_API_ENDPOINT,
    ASTRA_DB_KEYSPACE,
    ASTRA_DB_TOKEN_PROVIDER,
    IS_ASTRA_DB,
    LOCAL_DATA_API_ENDPOINT,
    LOCAL_DATA_API_KEYSPACE,
    LOCAL_DATA_API_TOKEN_PROVIDER,
)

from astrapy import DataAPIClient, Database
from astrapy.admin import parse_api_endpoint
from astrapy.constants import Environment
from astrapy.info import FindEmbeddingProvidersResult


def live_provider_info() -> FindEmbeddingProvidersResult:
    """
    Query the API endpoint `findEmbeddingProviders` endpoint
    for the latest information.

    This utility function uses the environment variables it can find
    to establish a target database to query.
    """

    database: Database
    if IS_ASTRA_DB:
        parsed = parse_api_endpoint(ASTRA_DB_API_ENDPOINT)
        if parsed is None:
            raise ValueError(
                "Cannot parse the Astra DB API Endpoint '{ASTRA_DB_API_ENDPOINT}'"
            )
        client = DataAPIClient(environment=parsed.environment)
        database = client.get_database(
            ASTRA_DB_API_ENDPOINT,
            token=ASTRA_DB_TOKEN_PROVIDER,
            keyspace=ASTRA_DB_KEYSPACE,
        )
    else:
        client = DataAPIClient(environment=Environment.OTHER)
        database = client.get_database(
            LOCAL_DATA_API_ENDPOINT,
            token=LOCAL_DATA_API_TOKEN_PROVIDER,
            keyspace=LOCAL_DATA_API_KEYSPACE,
        )

    database_admin = database.get_database_admin()
    response = database_admin.find_embedding_providers()
    return response
